import json
import os
import re
from typing import Any, Dict, List, Text, Tuple

import pymongo

from dialogues.bitod.src.knowledgebase.en_zh_mappings import BitodMapping
from dialogues.utils import constraint_list_to_dict

from .hk_mtr import MTR

value_mapping = BitodMapping()

is_mongo = True


def restaurants_en_US_booking(db, query, api_out_list=None):
    pre_api_return = {"user_name": query["user_name"], "number_of_people": query["number_of_people"], "time": query["time"]}
    api_out_list.remove("number_of_people")
    api_out_list.remove("time")
    api_out_list.remove("date")
    if "date" in query:
        pre_api_return["date"] = query["date"]
        del query["date"]
    del query["user_name"]
    query["max_num_people_book"] = {"$gte": query["number_of_people"]}
    del query["number_of_people"]

    # print("before: {}".format(query["time"]))
    temp = int(query["time"].split(":")[0])
    temp %= 12
    if "pm" in query["time"]:
        temp += 12
    mins = float(re.findall(r"[0-9][0-9]", query["time"].split(":")[1])[0]) / 60
    temp += mins
    # print(f"after: {temp}")
    query["time"] = temp
    query["open_time"] = {"$lte": query["time"]}
    query["close_time"] = {"$gte": query["time"]}
    del query["time"]
    res = list(db.find(query))

    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)

    if len(results) == 0:

        return (
            dict(
                Message="Sorry, the restaurant is not available given the booking time and number of people. The booking is failed."
            ),
            len(results),
            query,
        )
    else:
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return.update(pre_api_return)
        return api_return, len(results), query


def restaurants_zh_CN_booking(db, query, api_out_list=None):
    pre_api_return = {
        "user_name": query["user_name"],
        "number_of_people": query["number_of_people"],
        "time": query["time"],
        "date": query["date"],
    }
    api_out_list.remove("number_of_people")
    api_out_list.remove("time")
    api_out_list.remove("date")
    del query["date"]
    del query["user_name"]
    query["max_num_people_book"] = {"$gte": query["number_of_people"]}
    del query["number_of_people"]

    if "上午" in query["time"]:
        temp = query["time"].replace("上午", "")

    if "下午" in query["time"]:
        temp = query["time"].replace("下午", "")

    temp, m = temp.split(":")
    temp = int(temp)
    m = float(m)

    temp %= 12
    if "下午" in query["time"]:
        temp += 12

    mins = m / 60
    temp += mins
    # print(f"after: {temp}")
    query["time"] = temp
    query["open_time"] = {"$lte": query["time"]}
    query["close_time"] = {"$gte": query["time"]}
    del query["time"]
    res = list(db.find(query))

    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)

    if len(results) == 0:
        return dict(Message="对不起，预约失败。"), len(results), query
    else:
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return.update(pre_api_return)
        api_return = {value_mapping.en2zh_SLOT_MAP[k]: v for k, v in api_return.items()}
        return api_return, len(results), query


def hotels_en_US_booking(db, query, api_out_list=None):
    pre_api_return = {
        "user_name": query["user_name"],
        "number_of_rooms": query["number_of_rooms"],
        "start_month": query["start_month"],
        "start_day": query["start_day"],
        "number_of_nights": query["number_of_nights"],
    }
    api_out_list.remove("number_of_rooms")
    del query["user_name"]
    query["num_of_rooms"] = {"$gte": query["number_of_rooms"]}
    del query["number_of_rooms"]
    del query["start_month"]
    del query["start_day"]
    del query["number_of_nights"]

    res = list(db.find(query))

    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)

    if len(results) == 0:

        return (
            dict(Message="Sorry, the hotel is not available given the number of rooms. The booking is failed."),
            len(results),
            query,
        )
    else:
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return.update(pre_api_return)
        return api_return, len(results), query


def hotels_zh_CN_booking(db, query, api_out_list=None):
    pre_api_return = {
        "user_name": query["user_name"],
        "number_of_rooms": query["number_of_rooms"],
        "start_month": query["start_month"],
        "start_day": query["start_day"],
        "number_of_nights": query["number_of_nights"],
    }
    api_out_list.remove("number_of_rooms")
    del query["user_name"]
    query["num_of_rooms"] = {"$gte": query["number_of_rooms"]}
    del query["number_of_rooms"]
    del query["start_month"]
    del query["start_day"]
    del query["number_of_nights"]

    res = list(db.find(query))

    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)

    if len(results) == 0:

        return dict(Message="对不起，预约失败。"), len(results), query
    else:
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return.update(pre_api_return)
        api_return = {value_mapping.en2zh_SLOT_MAP[k]: v for k, v in api_return.items()}
        return api_return, len(results), query


def general_search_en_US(db, query, api_out_list=None):
    res = list(db.find(query).sort([("rating", pymongo.ASCENDING), ("_id", pymongo.DESCENDING)]))
    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)
    if len(results) == 0:
        return {}, len(results), query
    else:
        # print(res)
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return["available_options"] = len(results)

        if "price_per_night" in api_return:
            api_return["price_per_night"] = str(api_return["price_per_night"]) + " HKD"
        return api_return, len(results), query


def general_search_zh_CN(db, query, api_out_list=None):
    res = list(db.find(query).sort([("rating", pymongo.ASCENDING), ("_id", pymongo.DESCENDING)]))
    results = []
    for r in res:
        r["_id"] = str(r["_id"])
        results.append(r)
    if len(results) == 0:
        return {}, len(results), query
    else:
        # print(res)
        api_return = {k: results[-1][k] for k in api_out_list}
        api_return["available_options"] = len(results)

        if "price_per_night" in api_return:
            api_return["price_per_night"] = str(api_return["price_per_night"]) + "港币"
        api_return = {value_mapping.en2zh_SLOT_MAP[k]: v for k, v in api_return.items()}
        return api_return, len(results), query


def query_mongo(api_name, db, query, api_out_list=None):
    if api_name == "restaurants_en_US_booking":
        res, count, query = restaurants_en_US_booking(db, query, api_out_list)
    elif api_name == "hotels_en_US_booking":
        res, count, query = hotels_en_US_booking(db, query, api_out_list)
    elif api_name == "restaurants_zh_CN_booking":
        res, count, query = restaurants_zh_CN_booking(db, query, api_out_list)
    elif api_name == "hotels_zh_CN_booking":
        res, count, query = hotels_zh_CN_booking(db, query, api_out_list)
    elif "zh" in api_name:
        res, count, query = general_search_zh_CN(db, query, api_out_list)
    else:
        res, count, query = general_search_en_US(db, query, api_out_list)
    return res, count, query


def call_api(db, api_name, constraints: List[Dict[Text, Any]], lang=None) -> Tuple[Dict[Text, Any], int, dict]:
    api_name = value_mapping.r_en_API_MAP.get(api_name, api_name)

    # Canonicalization
    for slot, value in constraints[0].items():
        if isinstance(value, str) and (value in value_mapping.entity_map):
            constraints[0][slot] = value_mapping.entity_map[value]
        elif isinstance(value, dict):
            for k, v in value.items():
                if isinstance(v, str) and (v in value_mapping.entity_map):
                    constraints[0][slot][k] = value_mapping.entity_map[v]
                if isinstance(v, list):
                    constraints[0][slot][k] = [
                        value_mapping.entity_map[v_v] if v_v in value_mapping.entity_map else v_v for v_v in v
                    ]

    if api_name in [
        "restaurants_en_US_search",
        "restaurants_en_US_booking",
        "hotels_en_US_search",
        "hotels_en_US_booking",
        "attractions_en_US_search",
        "weathers_en_US_search",
        "餐馆查询",
        "餐馆预订",
        "宾馆查询",
        "宾馆预订",
        "景点查询",
        "天气查询",
    ]:

        if 'zh' in lang:
            api_name = value_mapping.en_zh_API_MAP.get(api_name, api_name)

        api_name = value_mapping.zh2en_API_MAP.get(api_name, api_name)
        constraints = [{value_mapping.zh2en_SLOT_MAP.get(k, k): v for k, v in constraints[0].items()}]

        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "apis", api_name + ".json"), "r") as file:
            api_schema = json.load(file)
        if constraints:
            all_provided_parameters = set.union(*[set(c) for c in constraints])
        else:
            all_provided_parameters = set()

        for parameter in api_schema["required"]:
            if parameter not in all_provided_parameters:
                raise ValueError(f"Parameter '{parameter}' is required but was not provided.")

        api_out_list = [slot["Name"] for slot in api_schema["output"]]

        if lang:
            if 'en' in lang:
                lang = 'en_US'
            api_db = db[re.sub(re.compile('(\w+_)\w{2}_\w{2}(_\w+)'), fr'\1{lang}\2', api_name)]
        else:
            api_db = db[api_name]

        res, count, query = query_mongo(api_name, api_db, constraint_list_to_dict(constraints), api_out_list)
        return res, count, query

    elif api_name in ["HKMTR_en", "HKMTR_zh", "香港地铁"]:

        if api_name == 'HKMTR_zh':
            api_name = '香港地铁'

        api_name = value_mapping.zh2en_API_MAP.get(api_name, api_name)
        constraints = [{value_mapping.zh2en_SLOT_MAP.get(k, k): v for k, v in constraints[0].items()}]

        source = constraint_list_to_dict(constraints)["departure"]
        target = constraint_list_to_dict(constraints)["destination"]
        if not lang:
            lang = api_name.split("_")[1]
        try:
            lang = lang[:2]
            mtr_dict = MTR(source=source, target=target, lang=lang)
        except Exception:
            return None, -1, None

        return mtr_dict, 1, None
    else:
        raise ValueError(f"API'{api_name}' is not available.")



import re


def replace_word(input, search, replace):
    def replace_method(match):
        if match.group(2) is None:
            return match.group()
        return match.group(2).replace(search, replace)

    expr = re.compile("('[^']*'|\"[^\"]*\")|({})".format(search))
    return re.sub(expr, replace_method, input)


s = "( movie ) production_country_or_area 年代 equal_to \" 年代中国大陆 \" , 年代 equal_to \" 2010 年代 \" 年代 production_country_or_area 年代 equal_to \" 年代中国大陆 \" ,"
output = replace_word(s, "年代", "decade")
print(output)



import copy
import json
import re
from collections import defaultdict

from ..main import Dataset


class Multiwoz24(Dataset):
    def __init__(self, name='multiwoz24'):
        super().__init__(name)

        self.state_re = re.compile('')
        self.knowledge_re = re.compile('')
        self.actions_re = re.compile('')

    def domain2api_name(self, domain):
        raise NotImplementedError

    def state2span(self, dialogue_state):
        raise NotImplementedError

    def span2state(self, state_span):
        raise NotImplementedError

    def process_state(self, state, **kwargs):
        belief = defaultdict(dict)
        for slot_value in state:
            assert slot_value['act'] == 'inform'

            assert len(slot_value['slots']) == 1
            domain_slot, value = slot_value['slots'][0]
            assert domain_slot.count('-') == 1
            domain, slot = domain_slot.split('-')

            assert isinstance(slot, str) and isinstance(value, str)

            # slot = slot.replace(' ', '-')

            # normalize so neural model can understand the words better
            # slot = slot.replace('pricerange', 'price-range')
            EXPERIMENT_DOMAINS = ["hotel", "train", "restaurant", "attraction", "taxi"]

            if value == 'none':
                value = None
            if domain not in EXPERIMENT_DOMAINS:
                value = None

            if value:
                # belief[domain][slot] = value
                belief[domain + '-' + slot] = value

        return belief

    def remove_extra_spaces(self, sent):
        for token in ['\t', '\n', ' ']:
            sent = sent.replace(token, ' ')
        sent = re.sub(r'\s{2,}', ' ', sent)
        return sent

    def process_agent_acts(self, acts_dict, ontology=None, **kwargs):
        acts = defaultdict(dict)
        for key, value in acts_dict.items():
            assert key.count('-') == 2
            domain, intent, slot = key.split('-', 2)

            # EXPERIMENT_DOMAINS = ["hotel", "train", "restaurant", "attraction", "taxi"]

            if slot in ['none', 'nooffer', 'ref']:
                continue
            if value in ['none']:
                value = None
            # if domain not in EXPERIMENT_DOMAINS:
            #     value = None

            if value:
                # belief[domain][slot] = value
                acts[key] = value

        return acts

    def state_dict2text(self, state, filtered=()):
        keys = list(state.keys())
        keys.sort()

        context = []
        for key in keys:
            value = state[key]
            if value in filtered:
                continue
            if value is None:
                continue
            if value == '':
                context.append(f'{key.replace("-", " ")} = " "')
            else:
                context.append(f'{key.replace("-", " ")} = " {value} "')

            context.append(',')

        if len(context) == 0:
            return 'null'

        return ' '.join(context).strip(', ')

    def process_data(self, args, root):
        splits = [split + '_dials_v2.json' for split in ['dev', 'train', 'test']]
        data_results = defaultdict(list)
        for split in splits:
            name = split.split('_')[0]
            data = json.load(open(root + split, 'r'))
            out_data = []
            for dlg in data:
                prev_belief_dict = dict()

                dlg_id = dlg['dialogue_idx']
                content = dlg['dialogue']
                turn_idx = 0

                while turn_idx < len(content):
                    turn = content[turn_idx]

                    EXPERIMENT_DOMAINS = ["hotel", "train", "restaurant", "attraction", "taxi"]
                    active_domain = turn['domain']

                    if active_domain not in EXPERIMENT_DOMAINS:
                        turn_idx += 1
                        continue

                    def get_lev_dict(belief_dict, prev_belief_dict):
                        lev_belief_dict = defaultdict(dict)
                        for domain_slot, value in belief_dict.items():
                            if domain_slot not in prev_belief_dict or prev_belief_dict[domain_slot] != value:
                                lev_belief_dict[domain_slot] = belief_dict[domain_slot]
                        return lev_belief_dict

                    belief_dict = self.process_state(turn['belief_state'])

                    user_utterance = self.remove_extra_spaces(turn["transcript"])
                    sys_utterance = self.remove_extra_spaces(turn["system_transcript"])

                    sys_acts = self.process_agent_acts(turn['system_acts'])

                    if not sys_utterance:
                        sys_utterance = '.'

                    # lev_dict = get_lev_dict(belief_dict, prev_belief_dict)
                    # lev_text = exp.state_dict2text(lev_dict)

                    belief_text = self.state_dict2text(belief_dict)
                    prev_belief_text = self.state_dict2text(prev_belief_dict)

                    acts = self.state_dict2text(sys_acts)

                    ex_id = f'{dlg_id}/{turn_idx}'

                    # context = ' '.join(['<state>', prev_belief_text, '<agent>', sys_utterance])
                    # question = ' '.join(['<user>', user_utterance])

                    input = ' '.join(
                        ["DST:", '<state>', prev_belief_text, '<endofstate>' '<acts>', acts, '<user>', user_utterance]
                    )

                    input = re.sub(r'\s{2,}', ' ', input)

                    out_string = '\t'.join([ex_id, input, belief_text])

                    # out_string = f'{ex_id}\t{prev_belief_text} ;{" " + sys_utterance if sys_utterance else ""}\t{user_utterance}\t{lev_text}'

                    out_data.append(out_string)

                    prev_belief_dict = copy.deepcopy(belief_dict)

                    turn_idx += 1

            data_results[name] = out_data

        return data_results['train'], data_results['dev'], data_results['test']

    def make_api_call(self, dialogue_state, api_name, **kwargs):
        raise NotImplementedError

    def compute_metrics(self, args, prediction_path, reference_path):
        raise NotImplementedError

    def postprocess_prediction(self, prediction, **kwargs):
        pass



import copy
import json
import os
import random
import re
from collections import OrderedDict, defaultdict

import dictdiffer
from datasets import load_metric
from tqdm import tqdm

from dialogues.utils import (
    clean_text,
    convert_to_int,
    is_at_least,
    is_equal_to,
    is_less_than,
    is_not,
    is_one_of,
    replace_word,
    zh2en_CARDINAL_MAP,
)

metric = load_metric("sacrebleu")


class Dataset(object):
    def __init__(self, name):

        # name of the dataset
        self.name = name

        # value mapping between different languages
        self.value_mapping = None

        # database
        self.db = None

        self.FAST_EVAL = False
        self.DEBUG = True

        # regex to extract belief state span from input
        self.state_re = re.compile('')

        # regex to extract knowledge span (api results) from input
        self.knowledge_re = re.compile('')

        # regex to extract dialogue history from input
        self.history_re = re.compile('')

        # regex to extract agent dialogue acts from input
        self.actions_re = re.compile('')

        self.system_token = ''
        self.user_token = ''

        if self.DEBUG:
            self.out_ser = open('out_ser.tsv', 'w')
            self.out_dst = open('out_dst.tsv', 'w')
            self.out_da = open('out_da.tsv', 'w')

    def domain2api_name(self, domain):
        """
        map domain name to api name used to query the database. these can be the same.
        :param domain: str
        :return: str
        """
        raise NotImplementedError

    def state2span(self, dialogue_state):
        """
        converts dictionary of dialogue state to a text span
        :param dialogue_state: dict
        :return: str
        """
        raise NotImplementedError

    def span2state(self, state_span):
        """
        converts dialogue state text span to a dictionary
        :param state_span: str
        :return: dict
        """
        raise NotImplementedError

    def update_state(self, lev, cur_state):
        """
        Updates cur_state according to the levenshtein state
        :param lev: dict
        :param cur_state: dict
        :return: dict
        """
        raise NotImplementedError

    def process_data(self, args):
        """
        converts raw dataset to the format accepted by genienlp
        each dialogue turn is broken down into 4 subtasks:
        Dialogue State Tracking (DST), API call decision (API) , Dialogue Act generation (DA), Response Generation (RG)
        :param args: dictionary of arguments passed to underlying data processor
        :return: three lists containing dialogues for each data split
        """
        raise NotImplementedError

    def make_api_call(self, dialogue_state, knowledge, api_name, **kwargs):
        """
        given dialogue state and api_name, compute the constraints for api call, make the call, and return the results in text form
        :param dialogue_state: dict
        :param knowledge: dict, keeping track returned api results
        :param api_name: str
        :param kwargs: additional args to pass to api call function
        :return: a dictionary of constraints used as input to api caller
        :return: text version of updated knowledge
        """
        raise NotImplementedError

    def compute_metrics(self, args, prediction_path, reference_path):
        """
        compare predictions vs gold and compute metrics. prediction file should contain model predictions for each subtask for each turn.
        reference file can be the original raw data file or a preprocessed version if it contains all needed info to calculate the required metrics
        :param args: dictionary of arguments passed to underlying evaluation code
        :param prediction_path: path to file containing model predictions
        :param reference_path: path to file containing gold values to compare predictions against
        :return: a dictionary with metric names as keys and their computed values (in percentage)
        """
        raise NotImplementedError

    def postprocess_prediction(self, prediction, **kwargs):
        """
        rule-based postprocessings done on model predictions
        :param prediction: str
        :param kwargs: additional args used for postprocessing
        :return: modified prediction
        """
        pass


class WOZDataset(Dataset):
    # Datasets annotated with slot-relation-values (does not have to be collected WOZ style)
    def __init__(self, name='woz'):
        super().__init__(name)

        # name of the dataset
        self.name = name

        # value mapping between different languages
        self.value_mapping = None

        self.db = None

        self.FAST_EVAL = False

        # regex to extract belief state span from input
        self.state_re = re.compile('<state> (.*?) <endofstate>')

        # regex to extract knowledge span (api results) from input
        self.knowledge_re = re.compile('<knowledge> (.*?) <endofknowledge>')

        # regex to extract dialogue history from input
        self.history_re = re.compile('<history> (.*?) <endofhistory>')

        # regex to extract agent dialogue acts from input
        self.actions_re = re.compile('<actions> (.*?) <endofactions>')

        # regex to user input from history
        self.user_re = re.compile('(?:USER|USER_ACTS): (.*?)(?:$|<)')

        # regex to system input from history
        self.system_re = re.compile('AGENT_ACTS: (.*?)(?:$|<)')

        self.system_token = 'AGENT_ACTS:'
        self.user_token = 'USER:'

    def domain2api_name(self, domain):
        """
        map domain name to api name used to query the database. these can be the same.
        :param domain: str
        :return: str
        """
        return domain

    def process_data(self, args):
        """
        converts raw dataset to the format accepted by genienlp
        each dialogue turn is broken down into 4 subtasks:
        Dialogue State Tracking (DST), API call decision (API) , Dialogue Act generation (DA), Response Generation (RG)
        :param args: dictionary of arguments passed to underlying data processor
        :return: three lists containing dialogues for each data split
        """
        if args.setting in ["en", "zh2en"]:
            path_train = ["data/en_train.json"]
            path_dev = ["data/en_valid.json"]
            path_test = ["data/en_test.json"]
        elif args.setting in ["zh", "en2zh"]:
            path_train = ["data/zh_train.json"]
            path_dev = ["data/zh_valid.json"]
            path_test = ["data/zh_test.json"]
        else:
            path_train = ["data/zh_train.json", "data/en_train.json"]
            path_dev = ["data/zh_valid.json", "data/en_valid.json"]
            path_test = ["data/zh_test.json", "data/en_test.json"]

        path_train = [os.path.join(args.root, p) for p in path_train]
        path_dev = [os.path.join(args.root, p) for p in path_dev]
        path_test = [os.path.join(args.root, p) for p in path_test]

        train, fewshot, dev, test = self.prepare_data(args, path_train, path_dev, path_test)
        return train, fewshot, dev, test

    def postprocess_prediction(self, prediction, **kwargs):
        """
        rule-based postprocessings done on model predictions
        :param prediction: str
        :param kwargs: additional args used for postprocessing
        :return: modified prediction
        """
        return prediction

    def update_state(self, lev, cur_state):
        """
        Updates cur_state according to the levenshtein state
        :param lev: dict
        :param cur_state: dict
        :return: dict
        """
        for api_name in lev:
            if api_name not in cur_state:
                cur_state[api_name] = lev[api_name]
            else:
                cur_state[api_name].update(lev[api_name])

    def compute_metrics(self, args, prediction_path, reference_path):
        """
        compare predictions vs gold and compute metrics. prediction file should contain model predictions for each subtask for each turn.
        reference file can be the original raw data file or a preprocessed version if it contains all needed info to calculate the required metrics
        :param args: dictionary of arguments passed to underlying evaluation code
        :param prediction_path: path to file containing model predictions
        :param reference_path: path to file containing gold values to compare predictions against
        :return: a dictionary with metric names as keys and their computed values (in percentage)
        """

        reference_data = {}
        for reference_file_path in reference_path.split("__"):
            with open(reference_file_path) as f:
                reference_data.update(json.load(f))

        with open(prediction_path) as f:
            predictions = json.load(f)

        if not args.setting:
            file = os.path.basename(reference_path)
            if 'zh' in file:
                args.setting = 'zh'
            else:
                args.setting = 'en'

        results = self.compute_result(predictions, reference_data)

        return results

    def construct_input(
        self,
        train_target,
        state=None,
        user_history=None,
        system_history=None,
        knowledge=None,
        actions=None,
        last_two_agent_turns=True,
        only_user_rg=True,
    ):
        if last_two_agent_turns and len(system_history) >= 2:
            history = [system_history[-2].replace('AGENT_ACTS:', 'AGENT_ACTS_PREV:'), system_history[-1], user_history[-1]]
        elif len(system_history) and len(user_history):
            history = [system_history[-1], user_history[-1]]
        elif len(user_history):
            history = [user_history[-1]]
        else:
            history = []

        history_text = " ".join(history)

        if train_target == 'dst':
            input_text = " ".join(
                [
                    "DST:",
                    "<state>",
                    state,
                    "<endofstate>",
                    "<history>",
                    history_text,
                    "<endofhistory>",
                ]
            )
        elif train_target == 'api':
            input_text = " ".join(
                [
                    "API:",
                    "<knowledge>",
                    knowledge,
                    "<endofknowledge>",
                    "<state>",
                    state,
                    "<endofstate>",
                    "<history>",
                    history_text,
                    "<endofhistory>",
                ]
            )
        elif train_target == 'da':
            input_text = " ".join(
                [
                    "DA:",
                    "<knowledge>",
                    knowledge,
                    "<endofknowledge>",
                    "<state>",
                    state,
                    "<endofstate>",
                    "<history>",
                    history_text,
                    "<endofhistory>",
                ]
            )

        elif train_target == 'rg':
            if only_user_rg:
                history_text = user_history[-1]

            input_text = " ".join(
                [
                    "RG:",
                    "<actions>",
                    actions,
                    "<endofactions>",
                    "<history>",
                    history_text,
                    "<endofhistory>",
                ]
            )

        return input_text

    def span2state(self, state_span):
        """
        converts dialogue state text span to a dictionary
        :param state_span: str
        :return: dict
        """
        # reverse direction of state2span function
        # converts text span to state dict

        state = defaultdict(dict)
        re_intent_spans = re.compile('\( (.*?) \)\s?(.*?)(?=$|\( )')
        re_srvs = re.compile('(\S*? \S*? " .*? "|\S*? #unknown)')

        matches = re_intent_spans.findall(state_span)

        for match in matches:
            intent, srv_span = match
            state[intent] = {}
            if intent in self.value_mapping.api_names:
                srv_matches = re_srvs.findall(srv_span)
                for srv in srv_matches:
                    if '#unknown' in srv:
                        continue
                    else:
                        try:
                            slot, relation, value = srv.split(' ', 2)
                        except Exception:
                            print(f'illegal syntax for slot-relation-values: {srv}')
                            continue

                    # remove " "
                    value = value[2:-2]
                    values = value.split(' | ')
                    values = [convert_to_int(val) for val in values]
                    state[intent][slot] = {"relation": relation, "value": values}

        return state

    def state2constraints(self, dict_data):
        # converts dialogue state to constraints canonical form
        constraints = {}
        for slot, r_v in dict_data.items():
            if r_v["value"] == ["don't care"] or r_v["value"] == ["不在乎"]:
                continue
            relation = r_v["relation"]
            values = r_v["value"]
            if relation != "one_of" and relation != self.value_mapping.en2zh_RELATION_MAP["one_of"]:
                values = values[0]
                if slot in [
                    'stars',
                    'rating',
                    'max_temp',
                    'min_temp',
                    'price_per_night',
                    'number_of_rooms',
                    'number_of_people',
                    'num_of_rooms',
                    'start_day',
                    'start_month',
                    'number_of_nights',
                ]:
                    values = convert_to_int(values, strict=False, word2number=True)
            if relation == "one_of" or relation == self.value_mapping.en2zh_RELATION_MAP["one_of"]:
                constraints[slot] = is_one_of(values)
            elif relation == "at_least" or relation == self.value_mapping.en2zh_RELATION_MAP["at_least"]:
                constraints[slot] = is_at_least(values)
            elif relation == "not" or relation == self.value_mapping.en2zh_RELATION_MAP["not"]:
                constraints[slot] = is_not(values)
            elif relation == "less_than" or relation == self.value_mapping.en2zh_RELATION_MAP["less_than"]:
                constraints[slot] = is_less_than(values)
            else:
                constraints[slot] = is_equal_to(values)
        return constraints

    def canonicalize_constraints(self, dict_data):
        # converts the constraints dictionary in the original data to canonical form
        constraints = {}
        for const in dict_data:
            for slot, values in const.items():
                relation = values[values.find(".") + 1 : values.find("(")]
                values = values[values.find("(") + 1 : -1]
                values = self.value_mapping.entity_map.get(values, values)

                if relation == "one_of" or relation == self.value_mapping.en2zh_RELATION_MAP["one_of"]:
                    values = values.split(" , ")
                else:
                    values = convert_to_int(values, word2number=True)
                if relation == "one_of" or relation == self.value_mapping.en2zh_RELATION_MAP["one_of"]:
                    constraints[slot] = is_one_of(values)
                elif relation == "at_least" or relation == self.value_mapping.en2zh_RELATION_MAP["at_least"]:
                    constraints[slot] = is_at_least(values)
                elif relation == "not" or relation == self.value_mapping.en2zh_RELATION_MAP["not"]:
                    constraints[slot] = is_not(values)
                elif relation == "less_than" or relation == self.value_mapping.en2zh_RELATION_MAP["less_than"]:
                    constraints[slot] = is_less_than(values)
                else:
                    constraints[slot] = is_equal_to(values)
        constraints = OrderedDict(sorted(constraints.items()))
        if constraints == {}:
            constraints = None
        return constraints

    def span2action(self, api_span):
        # reverse direction of state2span fuction
        # converts text span to state dict

        action = defaultdict(list)
        re_intent_spans = re.compile('\( (.+?) \) (.+?)(?=$|, \( )')
        re_asrs = re.compile('((?:\S+? ){3}" .+? "|\S*? \S+?(?: ,|$)|\S+?(?: ,|$))')

        matches = re_intent_spans.findall(api_span)

        for match in matches:
            intent, srv_span = match
            if intent in self.value_mapping.api_names:
                asr_matches = re_asrs.findall(srv_span)
                for asr in asr_matches:
                    slot, value, relation = 'null', 'null', 'null'
                    asr = asr.strip(' ,')
                    if '"' not in asr:
                        parts = asr.split(' ')
                        if len(parts) == 1:
                            act = parts[0]
                        else:
                            assert len(parts) == 2
                            act, slot = parts
                    else:
                        try:
                            act, slot, relation, value = asr.split(' ', 3)
                        except:
                            print(f'api_span: {api_span}, asr: {asr}')
                            continue

                    if value != 'null':
                        # remove " "
                        value = value[2:-2]
                        values = value.split(' | ')
                        values = [convert_to_int(val) for val in values]
                    else:
                        values = ['null']

                    action[intent].append({"act": act, "slot": slot, "relation": relation, "value": values})

        return action

    def action2span_for_single_intent(self, agent_actions, intent, setting):
        action_text = f'( {intent} ) '

        # sort based on act, then slot
        agent_actions = list(sorted(agent_actions, key=lambda s: (s['act'], s['slot'])))

        for action in agent_actions:
            act, slot, relation, values = action['act'], action['slot'], action['relation'], action['value']
            act = act.lower()

            if not act:
                # TODO: there are few annotation errors for act in RiSAWOZ, which makes "act" field empty
                #       this may cause unknown problems
                #       I'll fix it later. Current solution is just a workaround.
                continue
            orig_act = act
            if setting == 'zh':
                if act not in self.value_mapping.en2zh_ACT_MAP:
                    print(f'Encountered illegal act: {act}')
                    continue
                act = self.value_mapping.en2zh_ACT_MAP[act]

            values = [val for val in values if val != ""]
            if len(values):
                values = ' | '.join(map(str, values))
            else:
                values = 'null'

            if orig_act in [
                'notify_success',
                'notify_fail',
                'affirm',
                'request_more',
                'goodbye',
                'greeting',
                'thank_you',
                'negate',
                # for RiSAWOZ
                'bye',
                'general',
            ]:
                action_text += f'{act} , '
            # for RiSAWOZ
            elif orig_act in ['inform', 'recommend']:
                # TODO: annotation error; missing slot
                if not slot or values == 'null':
                    continue
                action_text += f'{act} {slot} {relation} " {values} " , '
            elif orig_act in ['request', 'request_update']:
                action_text += f'{act} {slot} , '
            # for RiSAWOZ
            elif orig_act in ['no-offer']:
                if not slot:
                    slot = 'null'
                action_text += f'{act} {slot} , '
            else:
                if orig_act in ['confirm']:
                    if not slot:
                        slot = 'null'
                    if not relation:
                        relation = 'null'
                else:
                    assert slot, action
                    assert relation, action

                action_text += f'{act} {slot} {relation} " {values} " , '

        action_text = action_text.strip(', ')

        return action_text

    def action2span(self, agent_actions, intents, setting):
        # compatible action2span for both BiToD and RiSAWOZ
        action_text = ''
        if isinstance(intents, list):
            pass
            # assert setting == "zh"  # temporary setting since translation is not ready yet
        else:
            intents = [intents]
        for intent in intents:
            intent_action_text = self.action2span_for_single_intent(agent_actions, intent, setting)
            action_text += intent_action_text
        return action_text.strip()  # retain the original output for BiToD

    def state2span(self, state):
        """
        converts dictionary of dialogue state to a text span
        :param dialogue_state: dict
        :return: str
        """

        if not state:
            return "null"
        span = ""

        # sort based on intent and then sort slots for each intent
        state = OrderedDict(sorted(state.items(), key=lambda s: s[0]))
        state = {k: OrderedDict(sorted(v.items(), key=lambda s: s[0])) for k, v in state.items()}

        def create_span(state, intent, slot):
            relation = state[intent][slot]["relation"]
            if isinstance(state[intent][slot]["value"], list):
                values = [str(value) for value in state[intent][slot]["value"]]
            else:
                # for RiSAWOZ
                values = [str(state[intent][slot]["value"])]
            values = sorted(values)
            values = " | ".join(values)
            span = f'{slot} {relation} " {values} " , '

            return span

        for intent in state:
            span += f"( {intent} ) "
            # check the required slots
            if intent not in self.value_mapping.required_slots:
                print(f'{intent} not in required slots!')
                continue
            if len(self.value_mapping.required_slots[intent]) > 0:
                for slot in self.value_mapping.required_slots[intent]:
                    if slot in state[intent]:
                        span += create_span(state, intent, slot)
                    else:
                        span += f"{slot} #unknown , "
            else:
                for slot in state[intent]:
                    span += create_span(state, intent, slot)
        return span.strip(', ')

    def compute_lev_span(self, previous_state, new_state, intent):
        Lev = f"( {intent} ) "
        if intent == "chat" or intent == "通用":
            return "null"
        old_state = copy.deepcopy(previous_state)
        if intent not in old_state:
            old_state[intent] = {}
        # TODO: annotation error of states in RiSAWOZ
        #       some states (not at begin of dialog) are empty
        #       I'll fix it later. Current solution is just a workaround.
        if intent not in new_state.keys():
            new_state[intent] = old_state[intent]

        for slot in new_state[intent]:
            if old_state[intent].get(slot) != new_state[intent].get(slot):
                relation = new_state[intent][slot]["relation"]
                if isinstance(new_state[intent][slot]["value"], list):
                    values = [str(value) for value in new_state[intent][slot]["value"]]
                else:
                    # for RiSAWOZ
                    values = [str(new_state[intent][slot]["value"])]
                values = " | ".join(values)
                Lev += f"{slot} {relation} \" {values} \" , "
        for slot in old_state[intent]:
            if slot not in new_state[intent]:
                print(f'slot: {old_state[intent][slot]} for intent: {intent} is missing in the new state')
                Lev += f"{slot} #unknown , "
        return Lev.strip(' ,')

    def span2knowledge(self, api_span):
        # reverse direction of knowledge2span fuction
        # converts knowledge span to knowledge dict

        knowledge = defaultdict(list)
        re_intent_spans = re.compile('\( (.*?) \)\s?(.*?)(?=$|\( )')
        re_svs = re.compile('(\S*? " .*? ")')

        matches = re_intent_spans.findall(api_span)

        for match in matches:
            intent, sv_span = match
            sv_matches = re_svs.findall(sv_span)
            for sv in sv_matches:
                try:
                    slot, value = sv.split(' ', 1)
                except Exception:
                    print(f'illegal syntax for slot-relation-values: {sv}')
                    exit(1)

                # remove " "
                value = value[2:-2]
                values = value.split(' | ')
                values = [convert_to_int(val) for val in values]
                knowledge[intent].append({"slot": slot, "value": values})

        return knowledge

    def knowledge2span(self, knowledge):
        if not knowledge:
            return 'null'

        # sort based on intent and then sort slots for each intent
        knowledge = OrderedDict(sorted(knowledge.items(), key=lambda s: s[0]))
        knowledge = {k: OrderedDict(sorted(v.items(), key=lambda s: s[0])) for k, v in knowledge.items()}

        knowledge_text = ""
        for intent, item in knowledge.items():
            knowledge_text += f"( {intent} ) "
            for slot, values in item.items():
                if slot not in self.value_mapping.skip_slots_for_kb:
                    if isinstance(values, list):
                        values_text = " | ".join(values)
                    else:
                        values_text = str(values)
                    if values_text == "":
                        values_text = "null"
                    knowledge_text += f"{slot} \" {values_text} \" , "
        return knowledge_text.strip(', ')

    def compute_da(self, preds, refs):
        da = 0.0
        for pred, ref in zip(preds, refs):
            if pred:
                pred = self.clean_value(pred)
                pred_dict = self.span2action(pred)

                ref = self.clean_value(ref)
                ref_dict = self.span2action(ref)

                if pred_dict == ref_dict:
                    da += 1
                else:
                    if self.DEBUG:
                        self.out_da.write(str(pred) + '\t' + str(ref) + '\t' + str(list(dictdiffer.diff(pred, ref))) + '\n')

        return da / len(preds) * 100

    def compute_ser(self, preds, act_values):
        ser = 0.0
        for pred, values in zip(preds, act_values):
            # remove emtpy slot values
            missing = False
            if len(values):
                for val in values:
                    if val in ['null', 'yes', 'no', 'true', 'false', '否', '是']:
                        continue
                    if str(val) not in pred:
                        missing = True
            if missing:
                ser += 1.0
                if self.DEBUG:
                    self.out_ser.write('\t'.join([pred, *values]) + '\n')
        return ser / len(preds) * 100

    def compute_dst_em(self, preds, golds):
        hit = 0
        for pred, gold in zip(preds, golds):
            pred_sets = self.convert_lists_to_set(pred)
            gold_sets = self.convert_lists_to_set(gold)

            if pred_sets == gold_sets:
                hit += 1
            else:
                if self.DEBUG:
                    self.out_dst.write(
                        str(pred_sets) + '\t' + str(gold_sets) + '\t' + str(list(dictdiffer.diff(pred_sets, gold_sets))) + '\n'
                    )

        return hit / len(preds) * 100

    def compute_success_rate(self, predictions, references):
        """
        Success:
        The system is able to offer the correct entities (e.g., restaurant name), provide the correct information (e.g., restaurant address),
        and confirm the booking information with the user before booking.

        Api call Accuracy:
        The predicted api call match the annotated api call.
        """

        total_dial = 0
        total_api_call = 0
        success_dial = 0
        correct_api_call = 0
        task_info = {}

        out_api = open('out_api.tsv', 'w')
        out_success = open('out_success.tsv', 'w')

        for dial_id in references:
            responses = ""
            total_dial += 1

            # api accuracy
            for api_name, constraints in references[dial_id]["API"].items():
                total_api_call += 1
                pred = predictions[dial_id]["API"].get(api_name)

                pred_sets = self.convert_lists_to_set_api(pred)
                constraints_sets = self.convert_lists_to_set_api(constraints)

                if pred_sets == constraints_sets:
                    correct_api_call += 1
                else:
                    out_api.write(
                        dial_id
                        + '\t'
                        + str(pred_sets)
                        + '\t'
                        + str(constraints_sets)
                        + '\t'
                        + str(list(dictdiffer.diff(pred_sets, constraints_sets)))
                        + '\n'
                    )
            api_acc = correct_api_call / total_api_call * 100

            # success
            out = dial_id + '\t'
            dial_success_flag = True
            for response in predictions[dial_id]["turns"].values():
                responses += response["response"][0] + " "
            responses = self.clean_value(responses)
            out += responses + '\t' + ';;;' + '\t'

            for intent in references[dial_id]["tasks"]:
                if intent not in task_info:
                    task_info[intent] = {"total": 0, "hit": 0, "success_rate": 0}
                task_success_flag = True
                task_info[intent]["total"] += 1

                for entity in (
                    references[dial_id]["tasks"][intent]["inform+offer"] + references[dial_id]["tasks"][intent]["confirmation"]
                ):
                    entity = self.clean_value(entity)
                    if entity in ['null', 'yes', 'no', 'true', 'false', '否', '是']:
                        continue
                    if str(entity) not in responses:
                        out += str(entity) + ' ; '
                        task_success_flag = False
                        break
                if task_success_flag:
                    task_info[intent]["hit"] += 1
                else:
                    dial_success_flag = False

            if dial_success_flag:
                success_dial += 1

            if out.split(';;;', 1)[1] != '\t':
                out_success.write(out + '\n')

        total_tasks = 0
        success_tasks = 0
        for task in task_info:
            task_info[task]["success_rate"] = task_info[task]["hit"] / task_info[task]["total"]
            total_tasks += task_info[task]["total"]
            success_tasks += task_info[task]["hit"]
        task_info["Averaged_task_success"] = success_tasks / total_tasks * 100
        success_rate = success_dial / total_dial * 100
        return success_rate, api_acc, task_info

    def clean_value(self, v, do_int=False):
        v = str(v)
        v = v.lower()
        v = v.strip()

        v = v.replace("，", ",")
        v = v.replace('..', '.')

        v = v.replace('；', ';')
        v = v.replace('。', '')

        # am, pm
        v = re.sub('(\d+)(?:[\.:](\d+))?\s?(?:pm )?(afternoon|in the afternoon|pm in the afternoon)', r'\1:\2 pm', v)
        v = re.sub('(\d+)(?:[\.:](\d+))?\s?(morning|in the morning|am in the morning)', r'\1:\2 am', v)
        v = re.sub('(\d+)(?:[\.:](\d+))?\s?(am|pm)', r'\1:\2 \3', v)

        v = re.sub('(\d+)\s?年\s?(\d+)\s?月\s?(\d+)', r'\1/\2/\3', v)
        if len(v) < 580:
            v = re.sub('(\d+)-?(\d+)?-?(\d+)?\s?\( (?:中国香港|中国大陆|韩国) \)?', r'\1/\2/\3', v)

        v = re.sub('(\d+)/0?(\d+)/0?(\d+)', r'\1/\2/\3', v)
        v = re.sub('(\d+)-0?(\d+)-0?(\d+)', r'\1/\2/\3', v)

        v = re.sub('(.*?) and (.*?) and (.*?)', r'\1,\2,\3', v)

        v = re.sub('(\d+)\.0', r'\1', v)

        # & --> and
        if self.name == 'bitod':
            v = re.sub(' [\&\/] ', ' and ', v)
        elif self.name == 'risawoz':
            v = re.sub(' \/ ', ',', v)

        v = re.sub('\s+', ' ', v)

        # remove extra dot in the end
        v = re.sub('(\d+)\.$', r'\1', v)
        v = re.sub('(\w+)\.$', r'\1', v)

        v = re.sub('(\w+)[。！？]$', r'\1', v)

        # 3rd of january --> januray 3
        v = re.sub('(\d+)(?:th|rd|st|nd) of (\w+)', r'\2 \1', v)

        # time consuming but needed step
        if not self.FAST_EVAL:
            for key, val in self.value_mapping.entity_map.items():
                key, val = str(key), str(val)
                if key in v:
                    v = v.replace(key, val)

        if do_int:
            v = convert_to_int(v, word2number=True)

        v = str(v)
        return v

    def convert_lists_to_set(self, state):
        new_state = copy.deepcopy(state)
        for i in state:
            for j in state[i]:
                for m, v in state[i][j].items():
                    if isinstance(v, list):
                        v = [self.clean_value(val, do_int=True) for val in v]
                        new_state[i][j][m] = set(v)
                    else:
                        new_state[i][j][m] = self.clean_value(v, do_int=True)
        return new_state

    def convert_lists_to_set_api(self, constraints):
        new_constraints = copy.deepcopy(constraints)
        if constraints:
            for k, v in constraints.items():
                if isinstance(v, dict):
                    for i, j in v.items():
                        if isinstance(j, list):
                            j = [self.clean_value(val, do_int=True) for val in j]
                            new_constraints[k][i] = set(j)
                        else:
                            new_constraints[k][i] = self.clean_value(j, do_int=True)
                else:
                    new_constraints[k] = self.clean_value(v, do_int=True)
        if new_constraints is None:
            new_constraints = {}
        return new_constraints

    def compute_result(self, predictions, reference_data):
        preds = []
        golds = []
        for dial_id in reference_data:
            pred_turn_id = 0
            for turn in reference_data[dial_id]["Events"]:
                if turn["Agent"] == "User":
                    pred_turn_id += 1

                    pred = predictions[dial_id]["turns"][str(pred_turn_id)]["state"]
                    # we record the string in genienlp instead of dict
                    # add fololoing for backward compatibility
                    if not isinstance(pred, dict):
                        pred = self.span2state(pred)
                    pred = {self.domain2api_name(k): v for k, v in pred.items()}
                    gold = turn["state"]

                    preds.append(pred)
                    golds.append(gold)

        jga = self.compute_dst_em(preds, golds)

        reference_task_success = defaultdict(dict)
        reference_act_values = []
        reference_actions = []
        reference_response = []
        predicted_response = []
        predicted_actions = []
        act_sets = set()
        for dial_id in reference_data:
            if dial_id not in reference_task_success:
                reference_task_success[dial_id]["tasks"] = {
                    self.value_mapping.zh_en_API_MAP.get(task["Task"], task["Task"]): {"inform+offer": [], "confirmation": []}
                    for task in reference_data[dial_id]["Scenario"]["WizardCapabilities"]
                }
                reference_task_success[dial_id]["API"] = {}
            pred_turn_id = 1
            user_requested_info = defaultdict(dict)
            confirm_info = defaultdict(dict)
            for turn in reference_data[dial_id]["Events"]:
                if turn["Agent"] == "User":
                    if not isinstance(turn["active_intent"], list):
                        # for compatibility of both BiTOD and RiSAWOZ
                        intent = [self.value_mapping.zh_en_API_MAP.get(turn["active_intent"], turn["active_intent"])]
                    else:
                        intent = turn["active_intent"]
                if turn["Agent"] == "Wizard":
                    if turn["Actions"] == "query":
                        if not isinstance(turn["API"], list):
                            # for compatibility of both BiTOD and RiSAWOZ
                            turn["API"] = [turn["API"]]
                            constraints = self.canonicalize_constraints(turn["Constraints"])
                        else:
                            # TODO: canonicalization for RiSAWOZ
                            constraints = turn["Constraints"]
                        for turn_api in turn["API"]:
                            turn_api = self.value_mapping.zh_en_API_MAP.get(turn_api, turn_api)
                            if constraints:
                                if turn_api in constraints.keys():
                                    # for RiSAWOZ: filter constraints with current API
                                    constraints = {k: v for k, v in constraints[turn_api].items()}
                                else:
                                    constraints = {k: v for k, v in constraints.items()}
                            reference_task_success[dial_id]["API"][turn_api] = constraints
                    else:
                        reference_response.append(self.clean_value(turn["Text"]))
                        act_values = set()
                        for item in turn["Actions"]:
                            if len(item["value"]):
                                if not isinstance(item["value"], list):
                                    item["value"] = [item["value"]]
                                for value in item["value"]:
                                    if not isinstance(value, list):
                                        value = [value]
                                    act_values.update(value)
                            act_values = set([self.clean_value(val) for val in act_values])
                        reference_act_values.append(act_values)

                        # for i in range(len(intent)):
                        #     # for RiSAWOZ: filter turn actions with current intent
                        #     turn_actions = (
                        #         [action for action in turn["Actions"] if action["domain"] == intent[i]]
                        #         if len(intent) > 1
                        #         else turn["Actions"]
                        #     )
                        #     reference_actions.append(
                        #         self.clean_value(
                        #             self.action2span(
                        #                 turn_actions, self.value_mapping.en_API_MAP.get(intent[i], intent[i]), 'en'
                        #             )
                        #         )
                        #     )

                        reference_actions.append(
                            self.clean_value(
                                self.action2span(
                                    turn["Actions"], [self.value_mapping.en_API_MAP.get(i, i) for i in intent], setting='en'
                                )
                            )
                        )

                        pred_rg = predictions[dial_id]["turns"][str(pred_turn_id)]["response"]
                        if isinstance(pred_rg, list):
                            assert len(pred_rg) == 1
                            pred_rg = pred_rg[0]
                        pred_rg = self.clean_value(pred_rg)
                        predicted_response.append(pred_rg)

                        pred_act = predictions[dial_id]["turns"][str(pred_turn_id)]["actions"]
                        if isinstance(pred_act, list):
                            assert len(pred_act) == 1
                            pred_act = pred_act[0]
                        pred_act = self.clean_value(pred_act)
                        predicted_actions.append(pred_act)

                        pred_turn_id += 1

                        # For each task, the last value for each slot are considered as final requested information from user
                        for action in turn["Actions"]:
                            if action["act"].lower() not in act_sets:
                                print(action["act"].lower())
                                act_sets.add(action["act"].lower())
                            trans_slot = action["slot"]
                            # should we include "recommend" ?
                            if (
                                (action["act"].lower() in ["inform", "offer"])
                                and (len(action["value"]) > 0)
                                and action["slot"] != "available_options"
                                and action["slot"] != "可用选项"
                            ):
                                for i in range(len(intent)):
                                    # for RiSAWOZ: filter turn actions with current intent
                                    if len(intent) <= 1 or action["domain"] == intent[i]:
                                        user_requested_info[intent[i]][trans_slot] = action["value"]
                            elif (action["act"] == "confirm") and (len(action["value"]) > 0):
                                # risawoz has no confirm act
                                assert len(intent) == 1
                                confirm_info[intent[0]][trans_slot] = action["value"]
            for intent, slot_values in user_requested_info.items():
                if intent in ["general"]:  # for RiSAWOZ
                    continue
                for values in slot_values.values():
                    reference_task_success[dial_id]["tasks"][intent]["inform+offer"] += values
            for intent, slot_values in confirm_info.items():
                if intent in ["general"]:  # for RiSAWOZ
                    continue
                for values in slot_values.values():
                    reference_task_success[dial_id]["tasks"][intent]["confirmation"] += values

        success_rate, api_acc, task_info = self.compute_success_rate(predictions, reference_task_success)

        bleu = self.compute_bleu(predicted_response, reference_response)
        ser = self.compute_ser(predicted_response, reference_act_values)
        da_acc = self.compute_da(predicted_actions, reference_actions)

        return OrderedDict(
            {
                "bleu": bleu,
                "ser": ser,
                "success_rate": success_rate,
                "api_acc": api_acc,
                "da_acc": da_acc,
                "jga": jga,
                "task_info": task_info,
            }
        )

    def postprocess_text(self, preds, labels):
        preds = [pred.strip() for pred in preds]
        labels = [[label.strip()] for label in labels]

        return preds, labels

    def compute_bleu(self, preds, labels):
        """
        preds = [pred1, pred2,...]
        labels = [label1, label2,...]
        """
        # Some simple post-processing
        preds, labels = self.postprocess_text(preds, labels)

        bleu = metric.compute(predictions=preds, references=labels)["score"]
        bleu = round(bleu, 4)

        return bleu

    def translate_slots_to_english(self, text, do_translate=True):
        if not do_translate:
            return text
        for key, val in self.value_mapping.translation_dict.items():
            text = replace_word(text, key, val)
        for key, val in self.value_mapping.zh_API_MAP.items():
            text = replace_word(text, key, val)
        for key, val in zh2en_CARDINAL_MAP.items():
            text = text.replace(f'" {key} "', f'" {val} "')
        return text

    def get_dials_sequential(self, args, dials):
        all_dial_ids = list(dials.keys())
        few_dials_file = os.path.join(args.root, f"data/{args.setting}_fewshot_dials_{args.fewshot_percent}.json")

        if not os.path.exists(few_dials_file):
            dial_ids = all_dial_ids[: int(len(all_dial_ids) * args.fewshot_percent // 100)]
            print(f"few shot for {args.setting}, dialogue number: {len(dial_ids)}")
            with open(few_dials_file, "w") as f:
                json.dump({"fewshot_dials": dial_ids}, f, indent=True)
        else:
            with open(few_dials_file) as f:
                dial_ids = json.load(f)["fewshot_dials"]
        few_dials = {dial_id: dials[dial_id] for dial_id in dial_ids}
        dials = {dial_id: dials[dial_id] for dial_id in all_dial_ids if dial_id not in dial_ids}

        return dials, few_dials

    def get_dials_balanced(self, args, dials):
        all_dial_ids = list(dials.keys())
        few_dials_file = os.path.join(args.root, f"data/{args.setting}_fewshot_dials_{args.fewshot_percent}_balanced.json")

        if not os.path.exists(few_dials_file):
            dialogue_dominant_domains = defaultdict(list)
            domain_turn_counts = defaultdict(int)
            for dial_id, dial in dials.items():
                turn_domains = defaultdict(int)
                dialogue_turns = dial["Events"]

                for turn in dialogue_turns:
                    if turn["Agent"] == "User":
                        active_intent = self.value_mapping.API_MAP[turn["active_intent"]]

                        turn_domains[self.translate_slots_to_english(active_intent, args.english_slots)] += 1
                        domain_turn_counts[self.translate_slots_to_english(active_intent, args.english_slots)] += 1

                max_domain = max(turn_domains, key=turn_domains.get)
                dialogue_dominant_domains[max_domain].append(dial_id)

            total = sum(list(domain_turn_counts.values()))
            fewshot_dials = []
            for (domain, count) in domain_turn_counts.items():
                num_fewshot = int(len(all_dial_ids) * (count / total) * (args.fewshot_percent / 100))
                fewshot_dials += dialogue_dominant_domains[domain][:num_fewshot]
            print(f"balanced few shot for {args.setting}, dialogue number: {len(fewshot_dials)}")
            print("turn statistics:")
            for (domain, count) in domain_turn_counts.items():
                print(domain, "comprises of", count, "or", int(100 * count / total + 0.5), "percent")
            print("few-shot turn statistics:")
            res_turn_counts = defaultdict(int)
            for dial_id in fewshot_dials:
                for turn in dials[dial_id]["Events"]:
                    if turn["Agent"] == "User":
                        active_intent = self.value_mapping.API_MAP[turn["active_intent"]]
                        res_turn_counts[self.translate_slots_to_english(active_intent, args.english_slots)] += 1
            total = sum(list(res_turn_counts.values()))
            for (domain, count) in res_turn_counts.items():
                print(domain, "comprises of", count, "or", int(100 * count / total + 0.5), "percent")

            with open(few_dials_file, "w") as f:
                json.dump({"fewshot_dials": fewshot_dials}, f, indent=True)
                few_dial_ids = fewshot_dials
        else:
            with open(few_dials_file) as f:
                few_dial_ids = json.load(f)["fewshot_dials"]
        few_dials = {dial_id: dials[dial_id] for dial_id in few_dial_ids}
        dials = {dial_id: dials[dial_id] for dial_id in all_dial_ids if dial_id not in few_dial_ids}

        return dials, few_dials

    def read_data(self, args, path_names):
        print(("Reading all files from {}".format(path_names)))

        max_history = args.max_history
        setting = args.setting
        # read files
        for path_name in path_names:
            with open(path_name) as file:
                dials = json.load(file)

                data = []
                for dial_id, dial in tqdm(dials.items()):
                    dialogue_turns = dial["Events"]

                    dialog_history = []
                    knowledge = defaultdict(dict)
                    prev_knowledge = "null"
                    prev_state = {}
                    count = 1
                    turn_id = 0

                    while turn_id < len(dialogue_turns):
                        turn = dialogue_turns[turn_id]

                        if turn["Agent"] == "User":
                            if not isinstance(turn["active_intent"], list):
                                # for compatibility of both BiTOD and RiSAWOZ
                                turn["active_intent"] = [turn["active_intent"]]

                            intents = [self.value_mapping.API_MAP[intent] for intent in turn["active_intent"]]
                            task = ' / '.join([self.value_mapping.zh2en_INTENT_MAP.get(intent, intent) for intent in intents])

                            # accumulate dialogue utterances
                            if args.use_user_acts:
                                # TODO: risawoz has multiple
                                assert len(intents) == 1
                                action_text = self.action2span(turn["Actions"], intents[0], setting)
                                action_text = clean_text(action_text, is_formal=True)
                                action_text = self.translate_slots_to_english(action_text, args.english_slots)
                                dialog_history.append("USER_ACTS: " + action_text)
                            else:
                                dialog_history.append("USER: " + clean_text(turn["Text"]))

                            if args.last_two_agent_turns and len(dialog_history) >= 4:
                                dial_hist = [dialog_history[-4].replace('AGENT_ACTS:', 'AGENT_ACTS_PREV:')] + dialog_history[
                                    -2:
                                ]
                            else:
                                dial_hist = dialog_history[-max_history:]

                            dialog_history_text = " ".join(dial_hist)
                            dialog_history_text_for_api_da = dialog_history_text

                            if args.only_user_rg:
                                dialog_history_text_for_rg = dial_hist[-1]
                            else:
                                dialog_history_text_for_rg = dialog_history_text

                            # convert dict of slot-values into text
                            state_text = self.state2span(prev_state)

                            current_state = {self.value_mapping.API_MAP[k]: v for k, v in turn["state"].items()}
                            current_state = OrderedDict(sorted(current_state.items(), key=lambda s: s[0]))
                            current_state = {
                                k: OrderedDict(sorted(v.items(), key=lambda s: s[0])) for k, v in current_state.items()
                            }

                            if args.gen_full_state:
                                targets = []
                                for intent in current_state.keys():
                                    targets.append(self.compute_lev_span({}, current_state, intent))
                                target = ' '.join(targets)
                            elif args.gen_lev_span:
                                # compute the diff between old state and new state
                                assert len(intents) == 1
                                target = self.compute_lev_span(prev_state, current_state, intents[0])
                            else:
                                assert len(intents) == 1
                                target = self.compute_lev_span({}, current_state, intents[0])

                            if not target:
                                target = 'null'

                            input_text = " ".join(
                                [
                                    "DST:",
                                    "<state>",
                                    self.translate_slots_to_english(state_text, args.english_slots),
                                    "<endofstate>",
                                    "<history>",
                                    dialog_history_text,
                                    "<endofhistory>",
                                ]
                            )

                            input_text = clean_text(input_text, True)

                            dst_data_detail = {
                                "dial_id": dial_id,
                                "task": task,
                                "turn_id": count,
                                "input_text": input_text,
                                "output_text": self.translate_slots_to_english(target, args.english_slots),
                                "train_target": "dst",
                            }

                            if args.detail:
                                dst_data_detail["active_intent"] = intents[0]
                                dst_data_detail["prev_state"] = prev_state
                                dst_data_detail["current_state"] = current_state

                            data.append(dst_data_detail)

                            # update last dialogue state
                            prev_state = current_state

                            turn_id += 1

                        elif turn["Agent"] == "Wizard":
                            # convert dict of slot-values into text
                            state_text = self.state2span(prev_state)

                            input_text = " ".join(
                                [
                                    "API:",
                                    "<knowledge>",
                                    self.translate_slots_to_english(prev_knowledge, args.english_slots),
                                    "<endofknowledge>",
                                    "<state>",
                                    self.translate_slots_to_english(state_text, args.english_slots),
                                    "<endofstate>",
                                    "<history>",
                                    dialog_history_text_for_api_da,
                                    "<endofhistory>",
                                ]
                            )

                            input_text = clean_text(input_text, True)

                            if turn["Actions"] == "query":
                                # do api call
                                # next turn is KnowledgeBase
                                assert dialogue_turns[turn_id + 1]["Agent"] == 'KnowledgeBase'
                                next_turn = dialogue_turns[turn_id + 1]

                                if int(next_turn["TotalItems"]) == 0:
                                    # TODO: risawoz has multiple intents
                                    prev_knowledge = f"( {intents[0]} ) Message = No item available."
                                else:
                                    for intent in intents:
                                        if intent in next_turn["Item"]:
                                            knowledge[intent].update(next_turn["Item"][intent])
                                        # general intent, 1 erroneous turn in zh_train
                                        elif self.name == 'risawoz':
                                            if intent != 'general':
                                                print(f'intent: {intent} not found in next_turn KB. skipping...')
                                            continue
                                        else:
                                            knowledge[intent].update(next_turn["Item"])

                                    prev_knowledge = self.knowledge2span(knowledge)

                                api_data_detail = {
                                    "dial_id": dial_id,
                                    "task": task,
                                    "turn_id": count,
                                    "input_text": input_text,
                                    "output_text": "yes",
                                    "train_target": "api",
                                }

                                data.append(api_data_detail)

                                # skip knowledge turn since we already processed it
                                turn_id += 2
                                turn = dialogue_turns[turn_id]

                            else:

                                # no api call
                                api_data_detail = {
                                    "dial_id": dial_id,
                                    "task": task,
                                    "turn_id": count,
                                    "input_text": input_text,
                                    "output_text": "no",
                                    "train_target": "api",
                                }

                                data.append(api_data_detail)

                            input_text = " ".join(
                                [
                                    "DA:",
                                    "<knowledge>",
                                    self.translate_slots_to_english(prev_knowledge, args.english_slots),
                                    "<endofknowledge>",
                                    "<state>",
                                    self.translate_slots_to_english(state_text, args.english_slots),
                                    "<endofstate>",
                                    "<history>",
                                    dialog_history_text_for_api_da,
                                    "<endofhistory>",
                                ]
                            )

                            input_text = clean_text(input_text, True)

                            target = clean_text(turn["Text"])

                            action_text = self.action2span(turn["Actions"], intents, setting)
                            action_text = clean_text(action_text, is_formal=True)
                            action_text = self.translate_slots_to_english(action_text, args.english_slots)

                            acts_data_detail = {
                                "dial_id": dial_id,
                                "task": task,
                                "turn_id": count,
                                "input_text": input_text,
                                "output_text": action_text,
                                "train_target": "da",
                            }
                            data.append(acts_data_detail)

                            input_text = " ".join(
                                [
                                    "RG:",
                                    "<actions>",
                                    action_text,
                                    "<endofactions>",
                                    "<history>",
                                    dialog_history_text_for_rg,
                                    "<endofhistory>",
                                ]
                            )

                            response_data_detail = {
                                "dial_id": dial_id,
                                "task": task,
                                "turn_id": count,
                                "input_text": input_text,
                                "output_text": target,
                                "train_target": "rg",
                            }
                            data.append(response_data_detail)

                            # update dialogue history
                            if args.use_natural_response:
                                output_text = target
                            else:
                                output_text = action_text

                            dialog_history.append("AGENT_ACTS: " + output_text)

                            turn_id += 1
                            count += 1

        return data

    def prepare_data(self, args, path_train, path_dev, path_test):

        # "en, zh, en&zh, en2zh, zh2en"
        data_train, data_fewshot, data_dev, data_test = None, None, None, None

        if 'valid' in args.splits:
            data_dev = self.read_data(args, path_dev)
        if 'test' in args.splits:
            data_test = self.read_data(args, path_test)
        if 'train' in args.splits:
            train_data = self.read_data(args, path_train)
            with open(path_train[0]) as file:
                dials = json.load(file)

            if args.sampling == "sequential":
                train_dials, few_dials = self.get_dials_sequential(args, dials)
            else:
                train_dials, few_dials = self.get_dials_balanced(args, dials)
            data_train, data_fewshot = [], []
            for data in train_data:
                if data['dial_id'] in train_dials:
                    data_train.append(data)
                else:
                    data_fewshot.append(data)

        if args.setting == "en_zh":
            if data_train:
                random.shuffle(data_train)
            if data_dev:
                random.shuffle(data_dev)
            if data_test:
                random.shuffle(data_test)

        return data_train, data_fewshot, data_dev, data_test



import logging

from pymongo import MongoClient

from ..main import WOZDataset
from .src.knowledgebase import api
from .src.knowledgebase.en_zh_mappings import RisawozMapping

logger = logging.getLogger(__name__)


class Risawoz(WOZDataset):
    def __init__(self, name='risawoz'):
        super().__init__(name)

        mongodb_host = 'mongodb+srv://bitod:plGYPp44hASzGbmm@cluster0.vo7pq.mongodb.net/risawoz?retryWrites=true&w=majority&ssl=true&ssl_cert_reqs=CERT_NONE'
        client = MongoClient(mongodb_host, authSource='admin')

        self.db = client["risawoz"]

        self.value_mapping = RisawozMapping()

        self._warnings = set()

    def domain2api_name(self, domain):
        return domain

    def make_api_call(
        self,
        dialogue_state,
        knowledge,
        api_names,
        src_lang='zh',
        dial_id=None,
        turn_id=None,
    ):
        constraints = {}
        for api_name in dialogue_state.keys():
            constraints[api_name] = self.state2constraints(dialogue_state[api_name])
        new_knowledge_text = 'null'

        try:
            result = api.call_api(self.db, api_names, constraints, src_lang, self.value_mapping)
            # remove _id
            # for api_name in result.keys():
            #     result[api_name].pop('_id', None)

        except Exception as e:
            logger.error(f'Error: {e}')
            logger.error(
                f'Failed API call with api_name: {str(api_names)}, constraints: {constraints}, for turn: {dial_id}/{turn_id}'
            )
        for api_name in result.keys():
            if int(len(result[api_name])) <= 0:
                warning = f'Message = No item available for api_name: {api_name}, constraints: {constraints}'
                if warning not in self._warnings:
                    logger.warning(warning + f', for turn: {dial_id}/{turn_id}')
                    self._warnings.add(warning)
                new_knowledge_text = f'( {api_name} ) Message = No item available.'
                # new_knowledge_text = 'null'
            else:
                knowledge.update(result)
                new_knowledge_text = self.knowledge2span(knowledge)

        return new_knowledge_text, constraints



import logging
import re

from pymongo import MongoClient

from ..main import WOZDataset
from .src.knowledgebase import api
from .src.knowledgebase.en_zh_mappings import BitodMapping

logger = logging.getLogger(__name__)


class Bitod(WOZDataset):
    def __init__(self, name='bitod'):
        super().__init__(name)

        self.value_mapping = BitodMapping()

        mongodb_host = 'mongodb+srv://bitod:plGYPp44hASzGbmm@cluster0.vo7pq.mongodb.net/bilingual_tod?retryWrites=true&w=majority&ssl=true&ssl_cert_reqs=CERT_NONE'
        client = MongoClient(mongodb_host, authSource='admin')
        database = client["bilingual_tod"]

        db = {"null": None}

        for domain in ['restaurants', 'hotels']:
            for lang in ['en_US', 'zh_CN']:
                db[f"{domain}_{lang}_booking"] = database[f"{domain}_{lang}"]
                db[f"{domain}_{lang}_search"] = database[f"{domain}_{lang}"]

        for domain in ['attractions', 'weathers']:
            for lang in ['en_US', 'zh_CN']:
                db[f"{domain}_{lang}_search"] = database[f"{domain}_{lang}"]

        self.db = db

    def domain2api_name(self, domain):
        # TODO: update
        return self.value_mapping.r_en_API_MAP.get(domain, domain)

    def make_api_call(self, dialogue_state, knowledge, api_names, src_lang='en', dial_id=None, turn_id=None):
        # bitod only does api call for the last (active) intent
        assert isinstance(api_names, list)
        api_name = api_names[-1]

        constraints = self.state2constraints(dialogue_state[api_name])

        count = 0
        processed_query = ''

        try:
            result, count, processed_query = api.call_api(self.db, api_name, constraints=[constraints], lang=src_lang)
        except Exception as e:
            logger.error(f'Error: {e}')
            logger.error(
                f'Failed API call with api_name: {str(api_name)}, constraints: {constraints}, processed_query: {processed_query}, for turn: {dial_id}/{turn_id}'
            )

        if int(count) <= 0:
            logger.warning(
                f'Message = No item available for api_name: {api_name}, constraints: {constraints},'
                f' processed_query: {processed_query}, for turn: {dial_id}/{turn_id}'
            )
            new_knowledge_text = f'( {api_name} ) Message = No item available.'
        else:
            # always choose the highest ranking results (so we have deterministic api results)
            knowledge[api_name].update(result)
            new_knowledge_text = self.knowledge2span(knowledge)

        return new_knowledge_text, {self.domain2api_name(api_name): constraints}

    def postprocess_prediction(self, prediction, knowledge=None, lang='en'):
        if re.search(rf'\( HKMTR {lang} \)', prediction):
            action_dict = self.span2action(prediction)
            domain = f'HKMTR {lang}'
            metro_slots = set(item['slot'] for item in action_dict[domain])
            for slot in ['estimated_time', 'price']:
                if knowledge and slot in knowledge[domain] and slot not in metro_slots:
                    action_dict[domain].append(
                        {'act': 'offer', 'slot': slot, 'relation': 'equal_to', 'value': [knowledge[domain][slot]]}
                    )

            prediction = self.action2span(action_dict[domain], domain, lang)

        if re.search(r'\( weathers search \)', prediction):
            action_dict = self.span2action(prediction)
            domain = 'weathers search'
            weather_slots = set(item['slot'] for item in action_dict[domain])
            for slot in ['max_temp', 'min_temp', 'weather', 'city']:
                if knowledge and slot in knowledge[domain] and slot not in weather_slots:
                    action_dict[domain].append(
                        {'act': 'offer', 'slot': slot, 'relation': 'equal_to', 'value': [knowledge[domain][slot]]}
                    )

            prediction = self.action2span(action_dict[domain], domain, lang)

        return prediction



import argparse
import json
import os

from BiToD.knowledgebase.en_zh_mappings import api_names, zh2en_ACT_MAP, zh2en_API_MAP, zh2en_RELATION_MAP, zh2en_SLOT_MAP
from BiToD.utils import span2action


class TemplateResponseGenerator(object):
    def __init__(self, lang, filename='human', template_dir=None):
        self.lang = lang
        if not template_dir:
            template_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')
        self.template_dir = template_dir
        templates = {}
        with open(os.path.join(self.template_dir, lang, f'{filename}.tsv')) as fin:
            for line in fin:
                id_, template = line.strip('\n').split('\t')
                # sanity check
                assert id_.count('@') == template.count('@'), (id_, template)
                templates[id_] = template
        self.templates = templates
        self.api_names = api_names

    def generate(self, acts):
        try:
            acts_dict = span2action(acts, self.api_names)
        except:
            print(f'Prediction contains slots not part of ontology (i.e. model hallucinated): {acts}')
            return 'None'
        # domains = list(processed_acts.keys())
        # keys.sort()

        context = []
        for domain_intent, asrvs in acts_dict.items():
            if self.lang == 'zh':
                domain_intent = zh2en_API_MAP[domain_intent]
                domain, intent = domain_intent.split('_')[0], domain_intent.split('_')[-1]
            else:
                domain, intent = domain_intent.split(' ')
            for item in asrvs:
                act, slot, relation, value = item['act'], item['slot'], item['relation'], item['value']
                if self.lang == 'zh':
                    act = zh2en_ACT_MAP[act]
                    slot = zh2en_SLOT_MAP[slot]
                    relation = zh2en_RELATION_MAP[relation]

                new_key = f'{domain}!!{act}!!{slot}'
                if len(value) != 1:
                    if relation == 'one_of':
                        print('No support for multiple values yet')
                        continue
                    else:
                        value = [' | '.join(value)]
                assert len(value) == 1
                value = str(value[0])

                if value != 'null':
                    new_key += '!!@'

                if self.lang == 'zh':
                    if new_key not in self.templates:
                        if slot == 'start_date':
                            new_key = new_key.replace('start_date', 'date')
                        # original dataset bug
                        if new_key == 'restaurants!!offer!!cuisine':
                            new_key = 'restaurants!!offer!!cuisine!!@'
                if new_key not in self.templates:
                    print('Failed to templatize')
                    print(new_key, acts)
                    continue
                template_text = self.templates[new_key].replace('@', value)

                context.append(template_text)

        return ' '.join(context)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--language', default='en')
    parser.add_argument('--template_name', default='human')
    parser.add_argument('--acts_file', required=True, help='path to file with input_acts')
    parser.add_argument('--processed_data_file', help='path to json file that has gold natural language responses')
    parser.add_argument('--output_file', required=True, help='path to output file with natural responses')
    args = parser.parse_args()

    templates = TemplateResponseGenerator(args.language, filename=args.template_name)

    # en_acts = '( attractions search )'
    #         ' offer address equal_to " 28 شان مونگ سړک ، کولوون " ,'
    #         ' offer name equal_to " اتاق چای گربه ها " ,'
    #         ' offer phone_number equal_to " +62 818-0919-0764 " ,'
    #         ' offer rating equal_to " 7 "'

    # zh_acts = '( 餐馆预订 ) ' \
    #        '确认 人数 等于 " 十一 " , ' \
    #        '确认 名字 等于 " 铃木屋 " , ' \
    #        '确认 时间 等于 " 下午7:40 " , ' \
    #        '确认 预订日期 等于 " 七月十二号 "'

    if args.processed_data_file:
        gold_responses = []
        with open(args.processed_data_file) as fin:
            data = json.load(fin)['data']
            for item in data:
                if 'response' in item:
                    gold_responses.append(item['response'])

    idx = 0
    with open(args.acts_file) as fin, open(args.output_file, 'w') as fout:
        for line in fin:
            parts = line.strip('\n').split('\t')
            if len(parts) == 1:
                sent = parts[0]
            elif len(parts) == 2:
                id_, sent = parts
            elif len(parts) == 3:
                id_, sent, gold = parts
            else:
                raise ValueError(f'Malformed input: {line}')

            if 'response' in id_:
                if args.processed_data_file:
                    gold = gold_responses[idx]
                    idx += 1
                new_sent = templates.generate(sent)
            else:
                new_sent = sent

            fout.write('\t'.join([id_, new_sent, gold]) + '\n')



import argparse
import json
import os

from dialogues.bitod.main import Bitod
from dialogues.utils import get_commit


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, default='dialogues/bitod/', help='code root directory')
    parser.add_argument("--data_dir", type=str, default="data", help="path to save original data, relative to root dir")
    parser.add_argument(
        "--detail", type=bool, default=False, help="whether to return dict annotations, used for data augmentation"
    )
    parser.add_argument(
        "--save_dir", type=str, default="data/preprocessed", help="path to save preprocessed data, relative to root dir"
    )
    parser.add_argument("--setting", type=str, default="en", help="en, zh, en_zh")

    parser.add_argument("--max_history", type=int, default=2)
    parser.add_argument("--splits", nargs='+', default=['train', 'valid', 'test'])
    parser.add_argument("--version", type=str, default='11')
    parser.add_argument("--fewshot_percent", type=int, default=0)
    parser.add_argument("--sampling", choices=["sequential", "balanced"], default="sequential")
    parser.add_argument("--use_user_acts", action='store_true')
    parser.add_argument("--gen_lev_span", action='store_true')
    parser.add_argument("--gen_full_state", action='store_true')
    parser.add_argument("--last_two_agent_turns", action='store_true')
    parser.add_argument("--english_slots", action='store_true')
    parser.add_argument("--use_natural_response", action='store_true')
    parser.add_argument("--only_user_rg", action='store_true')

    args = parser.parse_args()

    dataset = Bitod()

    if args.setting in ["en"]:
        path_train = ["data/en_train.json"]
        path_dev = ["data/en_valid.json"]
        path_test = ["data/en_test.json"]
    elif args.setting in ["zh"]:
        path_train = ["data/zh_train.json"]
        path_dev = ["data/zh_valid.json"]
        path_test = ["data/zh_test.json"]
    else:
        path_train = ["data/zh_train.json", "data/en_train.json"]
        path_dev = ["data/zh_valid.json", "data/en_valid.json"]
        path_test = ["data/zh_test.json", "data/en_test.json"]

    path_train = [os.path.join(args.root, p) for p in path_train]
    path_dev = [os.path.join(args.root, p) for p in path_dev]
    path_test = [os.path.join(args.root, p) for p in path_test]

    data_train, data_fewshot, data_dev, data_test = dataset.prepare_data(args, path_train, path_dev, path_test)

    args.commit = get_commit()

    save_dir = os.path.join(*[args.root, args.save_dir, f'{args.setting}_v{args.version}'])
    os.makedirs(save_dir, exist_ok=True)

    for (split, data) in zip(['train', 'fewshot', 'valid', 'test'], [data_train, data_fewshot, data_dev, data_test]):
        with open(os.path.join(save_dir, f"{split}.json"), "w") as f:
            if data:
                json.dump({"args": vars(args), "data": data}, f, indent=True, ensure_ascii=False)
                print(split, len(data))


if __name__ == "__main__":
    main()



import os
import re
import subprocess
from typing import Any, Dict, List, Text

from word2number import w2n

zh2en_CARDINAL_MAP = {
    "一": 1,
    "二": 2,
    "三": 3,
    "四": 4,
    "五": 5,
    "六": 6,
    "七": 7,
    "八": 8,
    "九": 9,
    "十": 10,
}
en2zh_CARDINAL_MAP = {v: k for k, v in zh2en_CARDINAL_MAP.items()}

zh2en_DAY_MAP = {
    "星期一": "Monday",
    "星期二": "Tuesday",
    "星期三": "Wednesday",
    "星期四": "Thursday",
    "星期五": "Friday",
    "星期六": "Saturday",
    "星期天": "Sunday",
}
en2zh_DAY_MAP = {v: k for k, v in zh2en_DAY_MAP.items()}
ENGLISH_MONTH_MAPPING = {
    '1': 'January',
    '2': 'February',
    '3': 'March',
    '4': 'April',
    '5': 'May',
    '6': 'June',
    '7': 'July',
    '8': 'August',
    '9': 'September',
    '10': 'October',
    '11': 'November',
    '12': 'December',
}
MONTH_MAPPING_EN = {
    k + ' month': v for k, v in sorted(ENGLISH_MONTH_MAPPING.items(), key=lambda item: int(item[0]), reverse=True)
}
MONTH_MAPPING_ZH = {
    v + '月': k + '月' for k, v in sorted(ENGLISH_MONTH_MAPPING.items(), key=lambda item: int(item[0]), reverse=True)
}


def convert_to_int(val, strict=False, word2number=False):
    val = str(val)
    if val.isdigit() and not val.startswith('0'):
        return int(val)
    elif word2number and len(val.split()) == 1:
        # elif word2number:
        try:
            num = w2n.word_to_num(val)
            return num
        except:
            if strict:
                return None
            else:
                return val
    else:
        if strict:
            return None
        else:
            return val


# very slow
def replace_word(input, search, replace):
    # return input.replace(search, replace)
    def replace_method(match):
        if match.group(2) is None:
            return match.group()
        return match.group(2).replace(search, replace)

    expr = re.compile(f"(\"[^\"]*\")|( {search}(?:$| ))")
    return re.sub(expr, replace_method, input)


def clean_text(text, is_formal=False):
    text = text.strip()
    text = re.sub(' +', ' ', text)
    text = re.sub('\\n|\\t', ' ', text)
    text = text.replace('，', ',')
    text = text.replace('，', ',')
    text = text.replace('？', '?')
    text = text.replace('！', '!')

    if not is_formal:
        text = text.replace('"', '')

    return text


def get_commit():
    directory = os.path.dirname(__file__)
    return (
        subprocess.Popen("cd {} && git log | head -n 1".format(directory), shell=True, stdout=subprocess.PIPE)
        .stdout.read()
        .split()[1]
        .decode()
    )


is_mongo = True


def is_equal_to(value):
    if is_mongo:
        return value  #
    else:
        return lambda x: x == value


def is_not(value):
    if is_mongo:
        return {"$ne": value}  #
    else:
        # return lambda x: x == value
        return lambda x: x != value


def contains_none_of(value):
    if is_mongo:
        return {"$nin": value}  #
    else:
        return lambda x: not any([e in x for e in value])


def is_one_of(value):
    if is_mongo:
        return {"$in": value}  #
    else:
        return lambda x: x in value


def is_at_least(value):
    if is_mongo:
        return {"$gte": value}  #
    else:
        return lambda x: x >= value


def is_less_than(value):
    if is_mongo:
        return {"$lt": value}  #
    else:
        return lambda x: x < value


def is_at_most(value):
    if is_mongo:
        return {"$lte": value}  #
    else:
        return lambda x: x <= value


def contain_all_of(value):
    return lambda x: all([e in x for e in value])


def contain_at_least_one_of(value):
    return lambda x: any([e in x for e in value])


def constraint_and(constraint1, constraint2):
    return lambda x: constraint1(x) and constraint2(x)


def constraint_list_to_dict(constraints: List[Dict[Text, Any]]) -> Dict[Text, Any]:
    result = {}
    for constraint in constraints:
        for name, constraint_function in constraint.items():
            # print(name, callable(constraint_function))
            if name not in result:
                result[name] = constraint_function
            else:
                result[name] = constraint_and(result[name], constraint_function)
    return result



import json
import os

import networkx as nx
import pydot

cur_dir = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(cur_dir, "mappings/mtr_mapping.json")) as f:
    MTR_map = json.load(f)

graph = pydot.graph_from_dot_file(os.path.join(cur_dir, "mappings/hongkong-mtr_with_eng.dot"))[0]
nx_graph = nx.drawing.nx_pydot.from_pydot(graph)


class key_dict(dict):
    def __getitem__(self, item):
        return item


name_to_en = key_dict()

name_to_fa = {
    "AsiaWorld-Expo": "نمایشگاه آسیای جهانی",
    "Airport": "فرودگاه",
    "Tsing Yi": "سینگ یو",
    "Kowloon": "کولوون",
    "Hong Kong": "هنگ کنگ",
    "Sunny Bay": "خلیج آفتابی",
    "Disneyland Resort": "استر استیشن دیزنی",
    "Diamond Hill": "هیل الماس",
    "Choi Wan": "چای وان",
    "Shun Tin": "شان تین",
    "Sau Mau Ping": "سو ماو پینگ",
    "Po Tat": "پو تات",
    "Po Lam": "پو لام",
    "Tuen Men South": "مردهای تیون جنوبی",
    "Tuen Men": "مردهای تیون",
    "Siu Hong": "سو هونگ",
    "Hung Shui Kiu": "هونگ شوی کیو",
    "Tin Shui Wai": "تین شوی وایی",
    "Long Ping": "لانگ پینگ",
    "Yuen Long": "یوین لانگ",
    "Kam Sheung Road": "جاده کام شونگ",
    "Tsuen Wan West": "تیون وان غربی",
    "Mei Foo": "می فو",
    "Nam Cheong": "نام چنگ",
    "Austin": "آستین",
    "East Tsim Sha Tsui": "شرق تیم شا تی",
    "Hung Hom": "هونگ هوم",
    "Ho Man Tin": "هو مان تین",
    "Ma Tau Wai": "ما تاو وایی",
    "To Kwa Wan": "تو کووا وان",
    "Kai Tak": "کای تاک",
    "Tai Wai": "تای وایی",
    "Che Kung Temple": "معبد چ کونگ",
    "Sha Tin Wai": "شا تین وایی",
    "City One": "شهر اول",
    "Shek Mun": "شیک مان",
    "Tai Shui Hang": "تای شوی هانگ",
    "Heng On": "هنگ آن",
    "Ma On Shan": "ما آن شان",
    "Wu Kai Sha": "و کای شا",
    "Kennedy Town": "شهر کندی",
    "HKU": "هانتون",
    "Sai Ying Pun": "سای ینگ پون",
    "Sheung Wan": "شونگ وان",
    "Central": "مرکزی",
    "Admirality": "Admirality (ادارایی)",
    "Wan Chai": "وان چی",
    "Causeway Bay": "خلیج کاوزウェی",
    "Tin Hau": "تین هاو",
    "Fortress Hill": "تپه قلعه",
    "North Point": "نقطه ی شمالی",
    "Quarry Bay": "خلیج کاوری",
    "Tai Koo": "تای کو",
    "Sai Wan Ho": "سای وان هو",
    "Shau Kei Wan": "شاو کی وان",
    "Heng Fa Chuen": "هنگ فا تشوان",
    "Chai Wan": "چای وان",
    "Whampoa": "ومپوا",
    "Yau Ma Tei": "یا ما تی",
    "Mong Kok": "منگ کوک",
    "Prince Edward": "پرنس ادوارد",
    "Shek Kip Mei": "شیک کیپ می",
    "Kowloon Tong": "کوولون تونگ",
    "Lok Fu": "لوک فو",
    "Wong Tai Sin": "وانگ تای سین",
    "Choi Hung": "چای هونگ",
    "Kowloon Bay": "خلیج کووان",
    "Ngau Tau Kok": "انگтау کوک",
    "Kwun Tong": "کوون تونگ",
    "Lam Tin": "لام تین",
    "Yau Tong": "یا تونگ",
    "Tiu Keng Leng": "تا کینگ لانگ",
    "Kwu Tung": "کو تونگ",
    "Exhibition": "نمایشگاه",
    "Mong Kok East": "منگ کوک شرق",
    "Sha Tin": "شا تین",
    "Fo Tan": "فو تان",
    "Racecourse": "میدان دونده",
    "University": "دانشگاه",
    "Tai Po Market": "بازار تای پو",
    "Tai Wo": "تای وو",
    "Fanling": "فَنَلینگ",
    "Sheung Shui": "شونگ شوی",
    "Lok Ma Chau": "لوک ما Chau",
    "Lo Wu": "لو و",
    "Ocean Park": "پارک اقیانوس",
    "Wong Chuk Hang": "وانگ چوک هانگ",
    "Lei Tung": "لی تونگ",
    "South Horizons": "افقی جنوبی",
    "Queen Mary Hospital": "بیمارستان ملکه مری",
    "Cyberport": "بندر سایبری",
    "Wah Fu": "ووه فو",
    "Tin Wan": "تین وان",
    "Aberdeen": "ابردین",
    "Tamar": "تامار",
    "Victoria Park": "پارک ویکتوریا",
    "Tseung Kwan O": "تسانگ کوان او",
    "Hang Hau": "هانگ هاو",
    "LOHAS Park": "پارک لوها",
    "Tsuen Wan": "تسو آن وان",
    "Tai Wo Hau": "تای و ها",
    "Kwai Hing": "کوای هنگ",
    "Kwai Fong": "کوای فنگ",
    "Lai King": "لای کینگ",
    "Lai Chi Kok": "لای چی کوک",
    "Cheung Sha Wan": "چانگ شا وان",
    "Sham Shui Po": "سام شوی پو",
    "Jordon": "اردن",
    "Tsim Sha Tsui": "تیم شا تی",
    "Tung Chung West": "تونگ چن غربی",
    "Tung Chung": "تونگ چینگ",
    "Tung Chung East": "تونگ چن شرقی",
    "Olympic": "اولمپیک",
    "Ngong Ping": "انگ پینگ",
}

name_to_zh = {
    "AsiaWorld-Expo": "博覽館",
    "Airport": "機場",
    "Tsing Yi": "青衣",
    "Kowloon": "九龍",
    "Hong Kong": "香港",
    "Sunny Bay": "欣澳",
    "Disneyland Resort": "迪士尼",
    "Diamond Hill": "鑽石山",
    "Choi Wan": "彩雲",
    "Shun Tin": "順天",
    "Sau Mau Ping": "秀茂坪",
    "Po Tat": "寶達",
    "Po Lam": "寶琳",
    "Tuen Men South": "屯門南",
    "Tuen Men": "屯門",
    "Siu Hong": "兆康",
    "Hung Shui Kiu": "洪水橋",
    "Tin Shui Wai": "天水圍",
    "Long Ping": "郎屏",
    "Yuen Long": "元郎",
    "Kam Sheung Road": "錦上路",
    "Tsuen Wan West": "荃灣西",
    "Mei Foo": "美孚",
    "Nam Cheong": "南昌",
    "Austin": "柯士甸",
    "East Tsim Sha Tsui": "尖東",
    "Hung Hom": "紅磡",
    "Ho Man Tin": "何文田",
    "Ma Tau Wai": "馬頭圍",
    "To Kwa Wan": "土瓜灣",
    "Kai Tak": "啟德",
    "Tai Wai": "大圍",
    "Che Kung Temple": "車公廟",
    "Sha Tin Wai": "沙田圍",
    "City One": "第一城",
    "Shek Mun": "石門",
    "Tai Shui Hang": "大水坑",
    "Heng On": "恆安",
    "Ma On Shan": "馬鞍山",
    "Wu Kai Sha": "烏溪沙",
    "Kennedy Town": "監尼地城",
    "HKU": "香港大學",
    "Sai Ying Pun": "西營盤",
    "Sheung Wan": "上環",
    "Central": "中環",
    "Admirality": "金鐘",
    "Wan Chai": "灣仔",
    "Causeway Bay": "銅鑼灣",
    "Tin Hau": "天后",
    "Fortress Hill": "炮台山",
    "North Point": "北角",
    "Quarry Bay": "鰂魚涌",
    "Tai Koo": "太古",
    "Sai Wan Ho": "西灣河",
    "Shau Kei Wan": "筲箕灣",
    "Heng Fa Chuen": "杏花邨",
    "Chai Wan": "柴灣",
    "Whampoa": "黃埔",
    "Yau Ma Tei": "油麻地",
    "Mong Kok": "旺角",
    "Prince Edward": "太子",
    "Shek Kip Mei": "石硤尾",
    "Kowloon Tong": "九龍塘",
    "Lok Fu": "樂富",
    "Wong Tai Sin": "黃大仙",
    "Choi Hung": "彩虹",
    "Kowloon Bay": "九龍灣",
    "Ngau Tau Kok": "牛頭角",
    "Kwun Tong": "觀塘",
    "Lam Tin": "藍田",
    "Yau Tong": "油塘",
    "Tiu Keng Leng": "調景嶺",
    "Kwu Tung": "古洞",
    "Exhibition": "會展",
    "Mong Kok East": "旺角東",
    "Sha Tin": "沙田",
    "Fo Tan": "火炭",
    "Racecourse": "馬場",
    "University": "大學",
    "Tai Po Market": "大埔墟",
    "Tai Wo": "太和",
    "Fanling": "粉嶺",
    "Sheung Shui": "上水",
    "Lok Ma Chau": "落馬洲",
    "Lo Wu": "羅湖",
    "Ocean Park": "海洋公園",
    "Wong Chuk Hang": "黃竹坑",
    "Lei Tung": "利東",
    "South Horizons": "海怡半島",
    "Queen Mary Hospital": "瑪麗醫院",
    "Cyberport": "數碼港",
    "Wah Fu": "華富",
    "Tin Wan": "田灣",
    "Aberdeen": "香港仔",
    "Tamar": "添馬",
    "Victoria Park": "維園",
    "Tseung Kwan O": "將軍澳",
    "Hang Hau": "坑口",
    "LOHAS Park": "康城",
    "Tsuen Wan": "荃灣",
    "Tai Wo Hau": "大窩口",
    "Kwai Hing": "葵芳",
    "Kwai Fong": "葵興",
    "Lai King": "荔景",
    "Lai Chi Kok": "荔枝角",
    "Cheung Sha Wan": "長沙灣",
    "Sham Shui Po": "生水埗",
    "Jordon": "佐敦",
    "Tsim Sha Tsui": "尖沙嘴",
    "Tung Chung West": "東涌",
    "Tung Chung": "東涌",
    "Tung Chung East": "東涌",
    "Olympic": "奧運",
    "Ngong Ping": "昂坪",
}
rev_name_to_zh = {v: k for k, v in name_to_zh.items()}

color = {
    "AE": "turquoise",
    "DR": "pink",
    "EK": "tan",
    "EW": "brown",
    "I": "blue",
    "KT": "green",
    "N": "black",
    "NS": "cyan",
    "SIE": "yellowgreen",
    "SIW": "mediumpurple",
    "TKO": "purple",
    "TW": "red",
    "TC": "orange",
    "NP": "orange",
}

color_to_en = key_dict()

color_to_zh = {
    "turquoise": "綠松石",
    "pink": "粉色的",
    "tan": "棕褐色",
    "brown": "棕色的",
    "blue": "藍色",
    "green": "綠色",
    "black": "黑色的",
    "cyan": "青色",
    "yellowgreen": "黃綠色",
    "mediumpurple": "中紫色",
    "purple": "紫色的",
    "red": "紅色的",
    "orange": "橘子",
}

color_to_fa = {
    "turquoise": "فیروزه",
    "pink": "صورتی",
    "tan": "برنزه",
    "brown": "قهوهای",
    "blue": "آبی",
    "green": "سبز",
    "black": "سیاه",
    "cyan": "فیروزهای",
    "yellowgreen": "سبز_زرد",
    "mediumpurple": "بنفش_متوسط",
    "purple": "بنفش",
    "red": "قرمز",
    "orange": "نارنجی",
}


name_line = {
    "AE": "Airport Express Line",
    "DR": "Disneyland Resort Line",
    "EK": "East Kowloon Line",
    "EW": "East West Line",
    "I": "Island Line",
    "KT": "Kwan Tong Line",
    "N": "Northern Line",
    "NS": "North South Line",
    "SIE": "South Island Line (East)",
    "SIW": "South Island Line (West)",
    "TKO": "Tseung Kwan O Line",
    "TW": "Tsuen Wan Line",
    "TC": "Tung Chung Line",
    "NP": "Tung Chung Line",
}

name_missing = {
    "AE1": {"label": "博覽館\rAsiaWorld-Expo"},
    "AE2": {"label": "機場\rAirport"},
    "AE3_TC5": {"label": "青衣\rTsing Yi"},
    "AE5_TC20": {"label": "香港\rHong Kong"},
    "DR1_TC4": {"label": "欣澳\rSunny Bay"},
    "DR2": {"label": "迪士尼\rDisneyland Resort"},
    "I1": {"label": "監尼地城\rKennedy Town"},
    "I2_SIW1": {"label": "香港大學\rHKU"},
    "I3": {"label": "西營盤\rSai Ying Pun"},
    "I4": {"label": "上環\rSheung Wan"},
    "I5_TW16": {"label": "中環\rCentral"},
    "I6_NS1_SIE1_TW15": {"label": "金鐘\rAdmirality"},
    "I7": {"label": "灣仔\rWan Chai"},
    "I8": {"label": "銅鑼灣\rCauseway Bay"},
    "I9": {"label": "天后\rTin Hau"},
    "I10": {"label": "炮台山\rFortress Hill"},
    "I11_TKO4": {"label": "北角\rNorth Point"},
    "I12_TKO5": {"label": "鰂魚涌\rQuarry Bay"},
    "I13": {"label": "西灣河\rSai Wan Ho"},
    "I14": {"label": "筲箕灣\rShau Kei Wan"},
    "I15": {"label": "杏花邨\rHeng Fa Chuen"},
    "I16": {"label": "柴灣\rChai Wan"},
    "SIE2": {"label": "海洋公園\rOcean Park"},
    "SIE3_SIW7": {"label": "黃竹坑\rWong Chuk Hang"},
    "SIE4": {"label": "利東\rLei Tung"},
    "SIE5": {"label": "海怡半島\rSouth Horizons"},
    "SIW2": {"label": "瑪麗醫院\rQueen Mary Hospital"},
    "SIW3": {"label": "數碼港\rCyberport"},
    "SIW4": {"label": "華富\rWah Fu"},
    "SIW5": {"label": "田灣\rTin Wan"},
    "SIW6": {"label": "香港仔\rAberdeen"},
    "TC21_TKO1": {"label": "添馬\rTamar"},
    "NS2_TKO2": {"label": "會展\rExhibition"},
    "TKO3": {"label": "維園\rVictoria Park"},
    "TC1": {"label": "東涌\rTung Chung West"},
    "TC2": {"label": "東涌\rTung Chung"},
    "TC3": {"label": "東涌\rTung Chung East"},
    "NP": {"label": "昂坪\rNgong Ping"},
}

for n in nx_graph.nodes():
    temp = {"name_en": "", "color": [], "line": []}
    if "label" in nx_graph.nodes[n]:
        temp["name_en"] = nx_graph.nodes[n]["label"].split("\\r")[1].replace('"', "")
    else:
        if n in name_missing:
            temp["name_en"] = name_missing[n]["label"].split("\r")[1]
    for id_ in n.split("_"):
        t = [x.isdigit() for x in id_]
        i = len(t)
        if True in t:
            i = [x.isdigit() for x in id_].index(True)
        id_ = id_[:i]

        temp["color"].append(color[id_])
        temp["line"].append(name_line[id_])

    nx_graph.nodes[n].update(temp)


G_MAPS = {'en': nx.Graph(), 'zh': nx.Graph(), 'fa': nx.Graph()}

for n in nx_graph.nodes():
    colors = nx_graph.nodes[n]["color"]

    for lang in ['en', 'zh', 'fa']:
        G = G_MAPS[lang]
        color_to_lang = eval(f"color_to_{lang}")
        name_to_lang = eval(f"name_to_{lang}")

        name = name_to_lang[nx_graph.nodes[n]["name_en"]]

        G.add_node('{name}')
        for c in colors:
            color = color_to_lang[c]
            G.add_node(f'{name}_{color}')
            G.add_edge(f'{name}', f'{name}_{color}', weight=0)
        for i in range(len(colors)):
            for j in range(i, len(colors)):
                G.add_edge(f'{name}_{color_to_lang[colors[i]]}', f'{name}_{color_to_lang[colors[j]]}', weight=1)

for s, t in nx_graph.edges():
    for lang in ['en', 'zh', 'fa']:
        G = G_MAPS[lang]
        color_to_lang = eval(f"color_to_{lang}")
        name_to_lang = eval(f"name_to_{lang}")
        s_name = name_to_lang[nx_graph.nodes[s]["name_en"]]
        s_colors = [color_to_lang[color] for color in nx_graph.nodes[s]["color"]]
        t_name = name_to_lang[nx_graph.nodes[t]["name_en"]]
        t_colors = [color_to_lang[color] for color in nx_graph.nodes[t]["color"]]
        for c in list(set(s_colors).intersection(set(t_colors))):
            G.add_edge(f'{s_name}_{c}', f'{t_name}_{c}', weight=1)


def MTR(source, target, lang="en"):
    new_source = MTR_map.get(source, source)
    new_target = MTR_map.get(target, target)
    G = G_MAPS[lang]
    shortest = nx.shortest_path(G, source=new_source, target=new_target)
    price = round((len(shortest) - 2) * 0.88 + 3.4, 2)
    time = (len(shortest) - 2) * 3

    # uncomment for shorten_path
    # if lang == 'en':
    #     str_ret = f"You will depart from {source} and arrive at {target}."
    #     return {"shortest_path": str_ret, "price": f"{price} HKD", "estimated_time": f"{time} mins"}
    # elif lang == 'zh':
    #     str_ret = f"你将从{source}出发，抵达{target}。"
    #     return {"最短路线": str_ret, "价格": f"{price}港币", "预估时间": f"{time}分钟"}
    if lang == "en":
        str_ret = ""
        str_ret += f"Take the {shortest[1].split('_')[1]} line of the {shortest[0]} station."
        for i in range(2, len(shortest) - 1):
            name, color = shortest[i].split("_")
            name_prev, color_prev = shortest[i - 1].split("_")
            if color != color_prev and name == name_prev:
                str_ret += f"Then change at {name} station from {color_prev} line to {color} line."
        str_ret += f"Get off the train at {shortest[-1]} station."
        return {"shortest_path": str_ret, "price": f"{price} HKD", "estimated_time": f"{time} mins"}
    elif lang == 'zh':
        str_ret = ""
        str_ret += f"请在{shortest[0]}站乘坐{shortest[1].split('_')[1]}线，"
        for i in range(2, len(shortest) - 1):
            name, color = shortest[i].split("_")
            name_prev, color_prev = shortest[i - 1].split("_")
            if color != color_prev and name == name_prev:
                str_ret += f"然后在{name}站换乘{color}线，"
        str_ret += f"最后在{shortest[-1]}站下车"
        return {"最短路线": str_ret, "价格": f"{price}港币", "预估时间": f"{time}分钟"}
    elif lang == 'fa':
        str_ret = ""
        str_ret += f"از خط {shortest[1].split('_')[1]} ایستگاه {shortest[0]} استفاده کنید."
        for i in range(2, len(shortest) - 1):
            name, color = shortest[i].split("_")
            name_prev, color_prev = shortest[i - 1].split("_")
            if color != color_prev and name == name_prev:
                str_ret += f"سپس در ایستگاه {name} از خط {color_prev} به خط {color} تغییر دهید."
        str_ret += f"در ایستگاه {shortest [-1]} از قطار پیاده شوید."
        return {"کوتاهترین_راه": str_ret, "قیمت": f"{price} HKD", "زمان_تخمین_زده_شده": f"{time}دقیقه "}



import copy
import re

from genienlp.data_utils.almond_utils import is_cjk_char


def call_api(db, api_names, constraints, lang, value_mapping, actions=None):
    knowledge = {}
    for api in api_names:
        api_en = value_mapping.zh2en_DOMAIN_MAP.get(api, api)
        # if api not in constraints:
        #     continue
        knowledge[api] = {}
        if api not in constraints:
            domain_constraints = {}
        else:
            domain_constraints = copy.deepcopy(constraints[api])

        domain_constraints = {k: value_mapping.en2canonical.get(v, v) for k, v in domain_constraints.items()}

        if api == 'car':
            if 'number_of_seats' in domain_constraints:
                domain_constraints['number_of_seats'] = {"$gte": int(domain_constraints['number_of_seats'])}

        db_name = f'{api_en}_{lang}'
        cursor = db[db_name].find(domain_constraints)
        domain_knowledge = []
        for matched in cursor:
            matched["_id"] = str(matched["_id"])
            # for key, val in matched.items():
            #     matched[key] = process_string(val)
            domain_knowledge.append(matched)
        if domain_knowledge:
            if actions:
                acts = actions[api]
                for item in domain_knowledge:
                    found = True
                    for slot, value in acts.items():
                        slot = slot.replace('.', '\uFF0E')
                        slot = slot.replace(' ', '_')
                        slot = slot.lower()
                        slot = value_mapping.zh2en_SLOT_MAP.get(slot, slot)
                        if slot not in item:
                            if slot == 'price' and api_en == 'car':
                                slot = 'price(ten_thousand_yuan)'
                            if slot == 'opening_hours' and api_en == 'restaurant':
                                slot = 'business_hours'
                        if slot not in item:
                            print(slot)
                        if item[slot] not in value:
                            found = False
                    if found:
                        knowledge[api] = item
                        break

            # if failed return first result
            # keep only the first result
            if not knowledge[api]:
                knowledge[api] = domain_knowledge[0]

            knowledge[api]["available_options"] = len(domain_knowledge)
    return knowledge


def tokenize_string(sentence):
    output = []
    i = 0
    while i < len(sentence):
        output.append(sentence[i])
        # skip space between cjk chars
        if (
            is_cjk_char(ord(sentence[i]))
            and i + 1 < len(sentence)
            and sentence[i + 1] == ' '
            and i + 2 < len(sentence)
            and is_cjk_char(ord(sentence[i + 2]))
        ):
            i += 1
        elif is_cjk_char(ord(sentence[i])) and i + 1 < len(sentence) and not is_cjk_char(ord(sentence[i + 1])):
            output.append(' ')
        elif not is_cjk_char(ord(sentence[i])) and i + 1 < len(sentence) and is_cjk_char(ord(sentence[i + 1])):
            output.append(' ')
        i += 1

    output = "".join(output)
    output = re.sub(r'\s{2,}', ' ', output)

    return output


def process_string(sentence, setting):
    if isinstance(sentence, bool):
        return str(sentence)
    if not isinstance(sentence, str):
        return sentence
    sentence = re.sub(r'\s{2,}', ' ', sentence)
    if setting == 'zh':
        sentence = ''.join(sentence.split())
    sentence = tokenize_string(sentence)

    return sentence



fa2en_CARDINAL_MAP = {
    "1": 1,
    "2": 2,
    "3": 3,
    "4": 4,
    "5": 5,
    "6": 6,
    "7": 7,
    "8": 8,
    "9": 9,
    "10": 10,
}
en2fa_CARDINAL_MAP = {v: k for k, v in fa2en_CARDINAL_MAP.items()}


en2fa_ACT_MAP = {
    'inform': 'اطلاع_دادن',
    'inform_intent': 'قصد_اطلاع_رسانی',
    'confirm': 'تایید',
    'offer': 'پیشنهاد',
    'request': 'درخواست',
    "request_more": "درخواست_بیشتر",
    "greeting": "تبریک",
    "thank_you": "متشکرم",
    "affirm": "تأیید_کردن",
    "negate": "نفی",
    "notify_success": "اطلاع_از_موفقیت",
    "notify_fail": "اطلاع_از_شکست",
    "goodbye": "خداحافظ",
    "request_update": "درخواست_به_روز_رسانی",
}
fa2en_ACT_MAP = {v: k for k, v in en2fa_ACT_MAP.items()}


en2fa_INTENT_MAP = {
    'inform': 'اطلاع_دادن',
    'confirm': 'تایید',
    'offer': 'پیشنهاد',
    'request': 'درخواست',
    "request_more": "درخواست_بیشتر",
    "greeting": "تبریک",
    "affirm": "تأیید_کردن",
    "notify_success": "اطلاع_از_موفقیت",
    "notify_fail": "اطلاع_از_شکست",
    "goodbye": "خداحافظ",
    "request_update": "درخواست_به_روز_رسانی",
}
fa2en_INTENT_MAP = {v: k for k, v in en2fa_INTENT_MAP.items()}

en2fa_SLOT_MAP = {
    "intent": "قصد",
    "name": "نام",
    "location": "محل",
    "rating": "رتبه_بندی",
    "type": "نوع",
    "address": "نشانی",
    "phone_number": "شماره_تلفن",
    "available_options": "گزینه_های_موجود",
    "ref_number": "شماره_مرجع",
    "price_per_night": "قیمت_شب",
    "user_name": "نام_کاربری",
    "number_of_rooms": "تعداد_اتاق_ها",
    "start_month": "شروع_ماه",
    "start_day": "روز_شروع",
    "start_date": "تاریخ_شروع",
    "number_of_nights": "تعداد_شبها",
    "price_level": "سطح_قیمت",
    "stars": "ستاره_ها",
    "cuisine": "غذا",
    "dietary_restrictions": "محدودیت_های_غذایی",
    "day": "روز",
    "city": "شهر",
    "max_temp": "حداکثر_دما",
    "min_temp": "دقیقه_دما",
    "weather": "آب_و_هوا",
    "description": "شرح",
    "departure": "عزیمت،_خروج",
    "destination": "مقصد",
    "number_of_people": "تعداد_مردم",
    "time": "زمان",
    "date": "تاریخ",
    "estimated_time": "زمان_تخمین_زده_شده",
    "shortest_path": "کوتاهترین_راه",
    "price": "قیمت",
}

fa2en_SLOT_MAP = {v: k for k, v in en2fa_SLOT_MAP.items()}


en2fa_RELATION_MAP = {"equal_to": "برابر", "not": "نابرابر", "less_than": "کمتر_از", "at_least": "حداقل", "one_of": "یکی_از"}
fa2en_RELATION_MAP = {v: k for k, v in en2fa_RELATION_MAP.items()}

fa2en_SPECIAL_MAP = {"بی_اهمیت": "don't care"}
en2fa_SPECIAL_MAP = {v: k for k, v in fa2en_SPECIAL_MAP.items()}


fa2en_API_MAP = {
    "جستجوی_رستوران": "restaurants_fa_XX_search",
    "رزرو_رستوران": "restaurants_fa_XX_booking",
    "جستجوی_هتل": "hotels_fa_XX_search",
    "رزرو_هتل": "hotels_fa_XX_booking",
    "جستجوی_جاذبه_ها": "attractions_fa_XX_search",
    "وضعیت_آب_و_هوا": "weathers_fa_XX_search",
    "مترو_هنگ_کنگ": "HKMTR_fa",
}
en2fa_API_MAP = {v: k for k, v in fa2en_API_MAP.items()}


# Mapping between api file names in kb/apis/* to a canonical name
fa2en_DOMAIN_MAP = {
    "جستجوی_رستوران": "restaurants search",
    "رزرو_رستوران": "restaurants booking",
    "جستجوی_هتل": "hotels search",
    "رزرو_هتل": "hotels booking",
    "جستجوی_جاذبه_ها": "attractions search",
    "وضعیت_آب_و_هوا": "weathers search",
    "مترو_هنگ_کنگ": "HKMTR en",
}
en2fa_DOMAIN_MAP = {v: k for k, v in fa2en_DOMAIN_MAP.items()}


en_API_MAP = {
    'chat': 'chat',
    "restaurants_en_US_search": "restaurants search",
    "restaurants_en_US_booking": "restaurants booking",
    "hotels_en_US_search": "hotels search",
    "hotels_en_US_booking": "hotels booking",
    "attractions_en_US_search": "attractions search",
    "weathers_en_US_search": "weathers search",
    "HKMTR_en": "HKMTR en",
}

fa_en_API_MAP = {
    "جستجوی_رستوران": "restaurants_en_US_search",
    "رزرو_رستوران": "restaurants_en_US_booking",
    "جستجوی_هتل": "hotels_en_US_search",
    "رزرو_هتل": "hotels_en_US_booking",
    "جستجوی_جاذبه_ها": "attractions_en_US_search",
    "وضعیت_آب_و_هوا": "weathers_en_US_search",
    "مترو_هنگ_کنگ": "HKMTR_en",
}
API_MAP = {}
API_MAP.update(en_API_MAP)
API_MAP.update(en2fa_API_MAP)
API_MAP.update({k: k for k, v in fa2en_API_MAP.items()})


# for cross lingual transfer
# mapping between slot values, not comprehensive, don't rely on it
# cur_dir = os.path.dirname(os.path.abspath(__file__))
# with open(os.path.join(cur_dir, "mappings/dict_en_fa.json")) as f:
#     en2fa_VALUE_MAP = json.load(f)
# fa2en_VALUE_MAP = {v: k for k, v in en2fa_VALUE_MAP.items()}


# maps entities to their canonicalized version; api expects canonicalized version
# note entities in original and preprocessed datasets are not canonicalized;
# cur_dir = os.path.dirname(os.path.abspath(__file__))
# entity_map = {}
# with open(os.path.join(cur_dir, "mappings/zh_entity_map.json")) as f:
#     entity_map.update(json.load(f))
# with open(os.path.join(cur_dir, "mappings/en_entity_map.json")) as f:
#     entity_map.update(json.load(f))
#
# reverse_entity_map = {v: k for k, v in entity_map.items()}



import argparse
import json

import requests

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--language', default='en')
    parser.add_argument('--model_url', default="http://localhost:8080/v1/models/nlp:predict")
    parser.add_argument('--acts_file', required=True, help='path to file with input_acts')
    parser.add_argument('--processed_data_file', help='path to json file that has gold natural language responses')
    parser.add_argument('--output_file', required=True, help='path to output file with natural responses')
    parser.add_argument('--batch_size', default=20)

    args = parser.parse_args()

    if args.processed_data_file:
        gold_responses = []
        with open(args.processed_data_file) as fin:
            data = json.load(fin)['data']
            for item in data:
                if 'response' in item:
                    gold_responses.append(item['response'])

    idx = 0
    batch = []
    batch_golds = []
    with open(args.acts_file) as fin, open(args.output_file, 'w') as fout:
        for line in fin:
            parts = line.strip('\n').split('\t')
            if len(parts) == 1:
                sent = parts[0]
            elif len(parts) == 2:
                id_, sent = parts
            elif len(parts) == 3:
                id_, sent, gold = parts
            else:
                raise ValueError(f'Malformed input: {line}')

            if 'response' in id_:
                if args.processed_data_file:
                    batch_golds.append(gold_responses[idx])
                    idx += 1
                batch.append({"context": sent, "question": ""})

                if len(batch) == args.batch_size:
                    request = json.dumps({"id": "null", "task": "bitod", "instances": batch})
                    response = requests.post(args.model_url, data=request)
                    preds = json.loads(response.text)["predictions"]
                    new_sents = [pred["answer"] for pred in preds]
                    for new_sent, gold in zip(new_sents, batch_golds):
                        fout.write('\t'.join([id_, new_sent, gold]) + '\n')
                    batch = []
                    batch_golds = []

        if len(batch):
            request = json.dumps({"id": "null", "task": "bitod", "instances": batch})
            response = requests.post(args.model_url, data=request)
            preds = json.loads(response.text)["predictions"]
            new_sents = [pred["answer"] for pred in preds]
            for new_sent, gold in zip(new_sents, batch_golds):
                fout.write('\t'.join([id_, new_sent, gold]) + '\n')



import argparse
import os

import ujson

from dialogues import Risawoz

risawoz = Risawoz()


def get_input_output(dic):
    dial_ids = []
    turn_ids = []
    inputs = []
    outputs = []
    entities = []
    for turn in dic["data"]:
        if turn['train_target'] not in ["dst", "da"]:
            continue

        dial_id = turn['dial_id']
        turn_id = turn['turn_id']
        input = turn['input_text']
        output = turn['output_text']

        dial_ids.append(dial_id)
        turn_ids.append(turn_id)
        inputs.append(input)
        outputs.append(output)

        ents = []

        if turn['train_target'] == 'dst':
            output = risawoz.span2state(output)
            for domain in output:
                for (slot, value) in output[domain].items():
                    if slot in ['the_most_suitable_people', 'metro_station']:
                        continue
                    for v in value['value']:
                        if v in ['null']:
                            continue
                        ents.append((slot, str(v)))
        else:
            output = risawoz.span2action(output)
            for domain in output:
                for item in output[domain]:
                    if item['slot'] in ['the_most_suitable_people', 'metro_station']:
                        continue
                    for v in item['value']:
                        if v in ['null']:
                            continue
                        ents.append((item['slot'], str(v)))

        entities.append(ents)

    return dial_ids, turn_ids, inputs, outputs, entities


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--directory", type=str, help='directory of files to check')
    parser.add_argument("--setting", type=str, default='zh')
    parser.add_argument("--splits", type=str, nargs='+', default=['train', 'valid', 'test'])
    parser.add_argument("--version", type=int, default=5)
    args = parser.parse_args()

    all_slots = list(risawoz.value_mapping.zh2en_SLOT_MAP.values())

    for split in args.splits:
        with open(os.path.join(args.directory, f"{split}_entity_check_{args.version}.tsv"), 'w+') as f:
            for (dial_id, turn_id, input, output, ents) in zip(
                *get_input_output(
                    ujson.load(open(os.path.join(args.directory, f'{args.setting}_{split}_v{args.version}.json')))
                )
            ):
                for (slot, ent) in ents:
                    if ent not in input:
                        f.write('\t'.join(map(str, [dial_id, turn_id, input, output, slot, ent])) + '\n')

    # for split in args.splits:
    #     outs_merged = ''
    #     for (dial_id, turn_id, input, output, ents) in zip(
    #             *get_input_output(
    #                 ujson.load(open(os.path.join(args.directory, f'{args.setting}_{split}_v{args.version}.json')))
    #             )
    #     ):
    #         outs_merged += output + ' '
    #
    # ignore_slots = []
    # for slot in all_slots:
    #     if slot not in outs_merged:
    #         print(f'slot {slot} is not present!')
    #         ignore_slots.append(slot)
    # print(ignore_slots)



import json
import os
from collections import OrderedDict

from dialogues.bitod.src.knowledgebase.hk_mtr import name_to_zh


class BitodMapping(object):
    def __init__(self):

        self.en2zh_RELATION_MAP = {"equal_to": "等于", "not": "非", "less_than": "少于", "at_least": "至少", "one_of": "其中之一"}
        self.zh2en_RELATION_MAP = {v: k for k, v in self.en2zh_RELATION_MAP.items()}

        self.zh2en_SPECIAL_MAP = {"不在乎": "don't care"}
        self.en2zh_SPECIAL_MAP = {v: k for k, v in self.zh2en_SPECIAL_MAP.items()}

        # TODO: missing
        self.DOMAIN_SLOT_MAP = {}

        self.zh2en_DOMAIN_MAP = {
            # currently untranslated
            "天气": "weather",
            "火车": "train",
            "电脑": "pc",
            "电影": "movie",
            "辅导班": "class",
            "汽车": "car",
            "餐厅": "restaurant",
            "酒店": "hotel",
            "旅游景点": "attraction",
            "飞机": "flight",
            "医院": "hospital",
            "电视剧": "tv",
            "通用": "general",
        }
        self.en2zh_DOMAIN_MAP = {v: k for k, v in self.zh2en_DOMAIN_MAP.items()}

        self.en2zh_ACT_MAP = {
            'inform': "通知",
            'inform_intent': '告知意图',
            # both affirm and confirm are mapped to 确认  smh...
            "affirm": "确认",
            'confirm': "确认",
            'offer': "报价",
            'request': "要求",
            "request_more": "要求更多",
            "greeting": "问候",
            "thank_you": "谢谢你",
            "negate": "否定",
            "notify_success": "通知成功",
            "notify_fail": "通知失败",
            "goodbye": "再见",
            "request_update": "请求更新",
        }
        self.zh2en_ACT_MAP = {v: k for k, v in self.en2zh_ACT_MAP.items()}

        self.zh2en_SLOT_MAP = {
            "意图": "intent",
            "名字": "name",
            "位置": "location",
            "评分": "rating",
            "类别": "type",
            "地址": "address",
            "电话": "phone_number",
            "可用选项": "available_options",
            "参考编号": "ref_number",
            "每晚价格": "price_per_night",
            "用户名": "user_name",
            "房间数": "number_of_rooms",
            "预订月": "start_month",
            "预订日": "start_day",
            "预订日期": "date",
            "预订天数": "number_of_nights",
            "价格范围": "price_level",
            "星级数": "stars",
            "菜品": "cuisine",
            "饮食限制": "dietary_restrictions",
            "日期": "day",
            "城市": "city",
            "最高温度": "max_temp",
            "最低温度": "min_temp",
            "天气": "weather",
            "描述": "description",
            "出发地": "departure",
            "目的地": "destination",
            "人数": "number_of_people",
            "时间": "time",
            # "日": "date",
            "预估时间": "estimated_time",
            "最短路线": "shortest_path",
            "价格": "price",
        }
        self.en2zh_SLOT_MAP = {v: k for k, v in self.zh2en_SLOT_MAP.items()}

        # for cross lingual transfer
        # mapping between slot values, not comprehensive, don't rely on it
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(cur_dir, "mappings/dict_en_zh.json")) as f:
            map_dict = json.load(f)

        en2zh_VALUE_MAP = name_to_zh
        en2zh_VALUE_MAP.update(map_dict)
        self.en2zh_VALUE_MAP = en2zh_VALUE_MAP

        zh2en_VALUE_MAP = {v: k for k, v in name_to_zh.items()}
        zh2en_VALUE_MAP.update({v: k for k, v in map_dict.items()})
        self.zh2en_VALUE_MAP = zh2en_VALUE_MAP

        # maps entities to their canonicalized version; api expects canonicalized version
        # note entities in original and preprocessed datasets are not canonicalized;
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        entity_map = {}
        with open(os.path.join(cur_dir, "mappings/zh_entity_map.json")) as f:
            entity_map.update(json.load(f))
        with open(os.path.join(cur_dir, "mappings/en_entity_map.json")) as f:
            entity_map.update(json.load(f))
        self.entity_map = entity_map

        self.reverse_entity_map = {v: k for k, v in self.entity_map.items()}

        cur_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(cur_dir, "mappings/zh2en_missing.json")) as f:
            self.zh2en_missing_MAP = json.load(f)
        with open(os.path.join(cur_dir, "mappings/en2zh_missing.json")) as f:
            self.en2zh_missing_MAP = json.load(f)

        # Mapping between api file names in kb/apis/* to a canonical name
        self.zh2en_API_MAP = {
            "餐馆查询": "restaurants_zh_CN_search",
            "餐馆预订": "restaurants_zh_CN_booking",
            "宾馆查询": "hotels_zh_CN_search",
            "宾馆预订": "hotels_zh_CN_booking",
            "景点查询": "attractions_zh_CN_search",
            "天气查询": "weathers_zh_CN_search",
            "香港地铁": "HKMTR_zh",
        }
        self.en2zh_API_MAP = {v: k for k, v in self.zh2en_API_MAP.items()}

        self.zh2en_INTENT_MAP = {
            "餐馆查询": "restaurants search",
            "餐馆预订": "restaurants booking",
            "宾馆查询": "hotels search",
            "宾馆预订": "hotels booking",
            "景点查询": "attractions search",
            "天气查询": "weathers search",
            "香港地铁": "HKMTR en",
        }
        self.en2zh_INTENT_MAP = {v: k for k, v in self.zh2en_INTENT_MAP.items()}

        self.en_API_MAP = {
            'chat': 'chat',
            "restaurants_en_US_search": "restaurants search",
            "restaurants_en_US_booking": "restaurants booking",
            "hotels_en_US_search": "hotels search",
            "hotels_en_US_booking": "hotels booking",
            "attractions_en_US_search": "attractions search",
            "weathers_en_US_search": "weathers search",
            "HKMTR_en": "HKMTR en",
        }
        self.r_en_API_MAP = {v: k for k, v in self.en_API_MAP.items()}

        self.zh_API_MAP = {
            'chat': 'chat',
            "restaurants_zh_CN_search": "restaurants search",
            "restaurants_zh_CN_booking": "restaurants booking",
            "hotels_zh_CN_search": "hotels search",
            "hotels_zh_CN_booking": "hotels booking",
            "attractions_zh_CN_search": "attractions search",
            "weathers_zh_CN_search": "weathers search",
            "HKMTR_zh": "HKMTR en",
        }

        self.zh_en_API_MAP = {
            "餐馆查询": "restaurants_en_US_search",
            "餐馆预订": "restaurants_en_US_booking",
            "宾馆查询": "hotels_en_US_search",
            "宾馆预订": "hotels_en_US_booking",
            "景点查询": "attractions_en_US_search",
            "天气查询": "weathers_en_US_search",
            "香港地铁": "HKMTR_en",
            "chat": "chat",
        }
        self.en_zh_API_MAP = {v: k for k, v in self.zh_en_API_MAP.items()}

        API_MAP = {}
        API_MAP.update(self.en_API_MAP)
        API_MAP.update(self.en2zh_API_MAP)
        API_MAP.update({k: k for k, v in self.zh2en_API_MAP.items()})
        API_MAP.update({'HKMTR zh': 'HKMTR zh'})

        self.API_MAP = API_MAP

        translation_dict = {
            **self.zh2en_API_MAP,
            **self.zh2en_DOMAIN_MAP,
            **self.zh2en_SLOT_MAP,
            **self.zh2en_ACT_MAP,
            **self.zh2en_RELATION_MAP,
            **self.zh2en_SPECIAL_MAP,
        }
        translation_dict["chat"] = "chat"
        self.translation_dict = OrderedDict(sorted(translation_dict.items(), key=lambda item: len(item[0]), reverse=True))

        required_slots = self.read_require_slots()
        self.required_slots = {self.API_MAP[k]: v for k, v in required_slots.items()}
        self.api_names = list(self.required_slots.keys())

        self.skip_slots_for_kb = ["type", "description", "类别", "描述"]

    def read_require_slots(self):
        require_slots = OrderedDict()
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        for fn in os.listdir(os.path.join(cur_dir, "apis")):
            with open(os.path.join(cur_dir, "apis", fn)) as f:
                ontology = json.load(f)
                api_name = fn.replace(".json", "")
                required = tuple(sorted(ontology["required"]))
                if 'zh' in api_name:
                    require_slots[api_name] = [self.en2zh_SLOT_MAP[val] for val in required]
                else:
                    require_slots[api_name] = required

        require_slots['HKMTR zh'] = require_slots['HKMTR_en']

        return require_slots



import argparse
import itertools
import json
import os
from collections import defaultdict
from contextlib import ExitStack
from pathlib import Path

import pymongo
import requests
from tqdm.autonotebook import tqdm

from dialogues.risawoz.main import Risawoz
from dialogues.risawoz.src.knowledgebase.api import call_api, process_string

dataset = Risawoz()


def read_json_files_in_folder(path):
    json_filename = [path + '/' + filename for filename in os.listdir(path) if '.json' in filename]
    with ExitStack() as stack:
        files = [stack.enter_context(open(fname)) for fname in json_filename]
        data = {}
        for i in range(len(files)):
            data[Path(json_filename[i]).stem] = json.load(files[i])
    return data


def build_db(db_json_path, api_map, setting, mongodb_host=""):
    raw_db = read_json_files_in_folder(db_json_path)
    if mongodb_host:
        db_client = pymongo.MongoClient(mongodb_host)
    else:
        db_client = pymongo.MongoClient()
    risawoz_db = db_client["risawoz"]
    for db in risawoz_db.list_collection_names():
        risawoz_db[db].drop()
    for domain in raw_db.keys():
        if api_map is None:
            col = risawoz_db[domain]
        else:
            col = risawoz_db[api_map[domain.split(f"_{setting}")[0]]]
        for i in range(len(raw_db[domain])):
            slot_list = list(raw_db[domain][i].keys())
            for s in slot_list:
                if "." in s:
                    # escape dot cause mongodb doesn't like '.' and '$' in key names
                    raw_db[domain][i][s.replace(".", "\uFF0E")] = raw_db[domain][i].pop(s)
        col.insert_many(raw_db[domain], ordered=True)
    return risawoz_db


def group_slot_values(actions, setting):
    # group slot values in user/system actions with same act-domain-slot prefix
    for i in range(len(actions)):
        actions[i][0] = actions[i][0].strip()  # strip spaces of ' inform '
    processed_actions = []
    grouped_actions = [list(v) for _, v in itertools.groupby(sorted(actions), lambda x: x[:3])]
    for group in grouped_actions:
        # TODO: the ''.join will cause mismatches between entities in input and output annotations
        # delete spaces and replace "" into []
        for i in range(len(group)):
            if group[i][3]:
                # keep space between cjk and other chars
                group[i][3] = [process_string(group[i][3], setting)]
            else:
                group[i][3] = []
        if len(group) == 1:
            group = group[0]  # squeeze
        else:
            group = group[0][:3] + [[action[3] for action in group]]
        processed_actions.append(group)
    return processed_actions


def build_user_event(turn, setting):
    event = {"Agent": "User"}
    # actions
    # TODO: handle domain information
    action_seq = ["act", "domain", "slot", "value"]
    actions = []
    processed_original_actions = group_slot_values(turn["user_actions"], setting)
    for action in processed_original_actions:
        event_action = {}
        for i in range(len(action)):
            if i == 1 and action[i]:
                action[i] = dataset.value_mapping.zh2en_DOMAIN_MAP.get(action[i], action[i].lower())
            elif i == 2 and action[i]:
                action[i] = dataset.value_mapping.zh2en_SLOT_MAP.get(action[i], action[i])
            elif i == 3 and action[i]:
                action[i] = process_string(action[i], setting)
            event_action[action_seq[i]] = action[i]
        if event_action["slot"] and event_action["value"]:
            event_action["relation"] = "equal_to"
        else:
            event_action["relation"] = ""
        event_action["active_intent"] = event_action["domain"]
        actions.append(event_action)
    event["Actions"] = actions
    # TODO: handle multiple active intents
    event["active_intent"] = [dataset.value_mapping.zh2en_DOMAIN_MAP.get(dom, dom) for dom in turn["turn_domain"]]
    event["state"] = defaultdict(dict)
    for ds, v in turn["belief_state"]["inform slot-values"].items():
        d, s = ds.split("-")[0], ds.split("-")[1]
        d = dataset.value_mapping.zh2en_DOMAIN_MAP.get(d, d)
        s = dataset.value_mapping.zh2en_SLOT_MAP.get(s, s)
        event["state"][d][s] = {"relation": "equal_to", "value": [process_string(v, setting)]}
    event["state"] = dict(event["state"])
    # event["Text"] = turn["user_utterance"]
    event["Text"] = process_string(turn["user_utterance"], setting)
    return event


def build_wizard_event(turn, setting, mode="normal"):
    assert mode in ["normal", "query"]
    event = {"Agent": "Wizard"}
    turn_domain_en = [dataset.value_mapping.zh2en_DOMAIN_MAP.get(d, d.lower()) for d in turn['turn_domain']]
    if mode == "query":
        event["Actions"] = "query"
        event["Constraints"] = defaultdict(dict)
        # event["Constraints_raw"] = defaultdict(dict)
        for ds, v in turn["belief_state"]["inform slot-values"].items():
            # only return matched result in the domains of current turn
            d, s = ds.split("-")
            d = dataset.value_mapping.zh2en_DOMAIN_MAP.get(d, d.lower())
            s = dataset.value_mapping.zh2en_SLOT_MAP.get(s, s)
            if d in turn_domain_en:
                event["Constraints"][d][s] = process_string(v, setting)
            # event["Constraints_raw"][d][s] = ''.join(v.split())
        # TODO: handle multiple APIs
        event["API"] = list(set([d for d in event["Constraints"].keys()]))
        if not event["API"]:
            event["API"] = turn_domain_en
    else:
        # actions
        action_seq = ["act", "domain", "slot", "value"]
        actions = []
        processed_original_actions = group_slot_values(turn["system_actions"], setting)
        for action in processed_original_actions:
            event_action = {}
            for i in range(len(action)):
                if i == 1 and action[i]:
                    action[i] = dataset.value_mapping.zh2en_DOMAIN_MAP.get(action[i], action[i].lower())
                elif i == 2 and action[i]:
                    action[i] = dataset.value_mapping.zh2en_SLOT_MAP.get(action[i], action[i])
                elif i == 3 and action[i]:
                    action[i] = process_string(action[i], setting)
                event_action[action_seq[i]] = action[i]
            event_action["relation"] = "equal_to" if event_action["slot"] and event_action["value"] else ""
            actions.append(event_action)
        event["Actions"] = actions

        # event["Text"] = turn["system_utterance"]
        event["Text"] = process_string(turn["system_utterance"], setting)
    return event


def build_kb_event(wizard_query_event, db, actions, expected_num_results, setting):
    event = {"Agent": "KnowledgeBase"}
    constraints = wizard_query_event["Constraints"]
    for d in constraints:
        constraints[d] = {k.replace(" ", "_"): v for k, v in constraints[d].items()}
    api_names = wizard_query_event["API"]
    knowledge = call_api(db, api_names, constraints, lang=setting, value_mapping=dataset.value_mapping, actions=actions)
    event["TotalItems"] = sum(item.get("available_options", 0) for api, item in knowledge.items())
    # if event["TotalItems"] == 0 and not expected_num_results == 0:
    # if event["TotalItems"] < expected_num_results and api_names != ['general']:
    for api, item in knowledge.items():
        if item.get("available_options", 0) < expected_num_results and api_names != ['general']:
            print('API call likely failed')
            knowledge = call_api(db, api_names, constraints, lang='zh', value_mapping=dataset.value_mapping, actions=actions)

    # for api, item in knowledge.items():
    #     for slot, val in item.items():
    #         knowledge[api][slot] = process_string(val, setting)
    event["Item"] = knowledge
    event["Topic"] = api_names
    return event


def build_dataset(original_data_path, db, setting):
    with open(original_data_path) as fin:
        data = json.load(fin)
    processed_data = {}
    for dialogue in tqdm(data):
        dialogue_id = dialogue["dialogue_id"]
        scenario = {
            "UserTask": dialogue.get("goal", ""),
            "WizardCapabilities": [
                {"Task": dataset.value_mapping.zh2en_DOMAIN_MAP.get(domain, domain.lower())} for domain in dialogue["domains"]
            ],
        }
        events = []
        turn_id = 0
        for turn in dialogue["dialogue"]:
            user_turn_event = build_user_event(turn, setting)
            if "db_results" in turn and turn["db_results"]:
                wizard_query_event = build_wizard_event(turn, setting, mode="query")
                wizard_normal_event = build_wizard_event(turn, setting, mode="normal")

                actions = defaultdict(defaultdict)
                for act in wizard_normal_event['Actions']:
                    domain, slot, value = act['domain'], act['slot'], act['value']
                    if slot:
                        actions[domain][slot] = value
                if setting == 'zh':
                    expected_num_results = int(turn["db_results"][0][len('数据库检索结果：成功匹配个数为') :])
                else:
                    expected_num_results = int(
                        turn["db_results"][0][len('Database search results: the number of successful matches is ') :]
                    )

                kb_event = build_kb_event(wizard_query_event, db, actions, expected_num_results, setting)
                user_turn_event['turn_id'] = turn_id
                wizard_query_event['turn_id'] = turn_id
                wizard_normal_event['turn_id'] = turn_id
                # del wizard_query_event['Constraints_raw']
                events += [user_turn_event, wizard_query_event, kb_event, wizard_normal_event]
            else:
                wizard_event = build_wizard_event(turn, setting)
                user_turn_event['turn_id'] = turn_id
                wizard_event['turn_id'] = turn_id
                events += [user_turn_event, wizard_event]

            turn_id += 1
        processed_data[dialogue_id] = {"Dialogue_id": dialogue_id, "Scenario": scenario, "Events": events}
    return processed_data


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--root", type=str, default='dialogues/risawoz/', help='code root directory')
    parser.add_argument("--data_dir", type=str, default="data/original/", help="path to original data, relative to root dir")
    parser.add_argument("--save_dir", type=str, default="data/", help="path to save preprocessed data, relative to root dir")
    parser.add_argument("--setting", type=str, default="en", help="en, zh, en_zh")
    parser.add_argument("--splits", nargs='+', default=['train', 'valid', 'test'])

    args = parser.parse_args()

    mongodb_host = "mongodb://localhost:27017/"

    # uncomment to build db
    risawoz_db = build_db(
        db_json_path=os.path.join(*[args.root, f'database/db_{args.setting}']),
        api_map=None,
        setting=args.setting,
        mongodb_host=mongodb_host,
    )

    # download original RiSAWOZ dataset
    original_data_path = os.path.join(*[args.root, args.data_dir])
    for split in args.splits:
        if not os.path.exists(os.path.join(original_data_path, f"{args.setting}_{split}.json")):
            os.makedirs(original_data_path, exist_ok=True)
            print(f"{split} set is not found, downloading...")
            if split == "valid":
                data_url = "https://huggingface.co/datasets/GEM/RiSAWOZ/resolve/main/dev.json"
            else:
                data_url = f"https://huggingface.co/datasets/GEM/RiSAWOZ/resolve/main/{split}.json"
            with open(f"{original_data_path}/{args.setting}_{split}.json", 'wb') as f:
                f.write(requests.get(data_url).content)

    processed_data_path = os.path.join(*[args.root, args.save_dir])
    for split in args.splits:
        print(f"processing {split} data...")
        processed_data = build_dataset(
            os.path.join(original_data_path, f"{args.setting}_{split}.json"), risawoz_db, args.setting
        )
        # save converted files in JSON format
        with open(f"{processed_data_path}/{args.setting}_{split}.json", 'w') as f:
            json.dump(processed_data, f, ensure_ascii=False, indent=4)



import argparse
import json
import os

from dialogues import Risawoz

parser = argparse.ArgumentParser()

parser.add_argument('--input_db_dir', default='dialogues/risawoz/database/db_zh')
parser.add_argument('--output_db_dir', default='dialogues/risawoz/database/db_in2')
parser.add_argument('--experiment')

args = parser.parse_args()


dataset = Risawoz()
value_mapping = dataset.value_mapping

seen = set()
out = open('./out_ent.tsv', 'w')

all_replacements = {}

all_replacements.update({'否': r'no'})

all_replacements.update({'当天 (\d{2}:\d{2})': r'\1 today'})

all_replacements.update({'(\d+\.?\d?)元': r'\1 yuan'})

all_replacements.update({'(\d{4}-\d{2}-\d{2})\(韩国\)': r'\1(South Korea)'})
all_replacements.update({'(\d{4}-\d{2}-\d{2})\(中国香港\)': r'\1(Hong Kong, China)'})
all_replacements.update({'(\d{4}-\d{2}-\d{2})\(日本\)': r'\1(Japan)'})
all_replacements.update({'(\d{4}-\d{2}-\d{2})\(美国\)': r'\1(USA)'})
all_replacements.update({'(\d{4}-\d{2}-\d{2})\(中国台湾\)': r'Taiwan, China'})
all_replacements.update({'(\d{4}-\d{2}-\d{2})\(中国大陆\)': r'Mainland, China'})

all_replacements.update({'周一至周日 (\d{2}:\d{2}-\d{2}:\d{2})': r'Monday to Sunday \1'})
all_replacements.update({'周一至周日 (\d{2}:\d{2}-\d{2}:\d{2}) (\d{2}:\d{2}-\d{2}:\d{2})': r'Monday to Sunday \1 \2'})


all_replacements.update({'(\d{4}/\d{1,2}/\d{1,2})': r'\1'})


os.makedirs(args.output_db_dir, exist_ok=True)

dicti = dict()

for file in os.listdir(args.input_db_dir):
    domain = file.split('_', 1)[0]
    if domain.startswith('.'):
        continue
    # if domain not in value_mapping.en2zh_DOMAIN_MAP.keys():
    #     continue
    db_info = json.load(open(os.path.join(args.input_db_dir, file)))

    translated_db_info = []

    if isinstance(db_info, dict):
        dicti[domain] = sorted(db_info.keys())
        # print(f'domain: {domain}, keys: {db_info.keys()}')
    else:
        dicti[domain] = sorted(db_info[0].keys())
        # print(f'domain: {domain}, keys: {db_info[0].keys()}')

    for item in db_info:
        translated_item = {}
        # for slot, value in item.items():
        #     pass
        # if slot not in value_mapping.zh2en_SLOT_MAP:
        #     if slot not in seen:
        #         print(slot)
        #         seen.add(slot)
        #     continue
        #
        # def find_new(value):
        #     value = str(value)
        #     for in_pattern, out_pattern in all_replacements.items():
        #         if re.fullmatch(in_pattern, value):
        #             new_value = re.sub(in_pattern, out_pattern, value)
        #             return new_value
        #
        #     return None

        # is_list = True
        # if not isinstance(value, list):
        #     is_list = False
        #     value = [value]
        # new_value = []
        # for val in value:
        #
        #     if val in [True, False]:
        #         new_value.append(val)
        #     elif val in value_mapping.zh2en_VALUE_MAP:
        #         new_value.append(value_mapping.zh2en_VALUE_MAP[val])
        #     elif val in value_mapping.zh2en_missing_MAP:
        #         new_value.append(value_mapping.zh2en_missing_MAP[val])
        #     elif find_new(val):
        #         new_value.append(find_new(val))
        #     else:
        #         if (
        #             val not in seen
        #             and not str(val).startswith('http:')
        #             and not re.fullmatch('[A-Z0-9]+', str(val))
        #             and not re.fullmatch('[\d\.:\-]+', str(val))
        #             and val not in [True, False]
        #         ):
        #             print(val)
        #             out.write(str(val) + '\n')
        #             seen.add(val)
        #
        #         if isinstance(val, str) and val.endswith('。') and slot == '特点':
        #             val = val[:-1]
        #
        #         new_value.append(val)
        #
        # value = new_value
        #
        # if not is_list:
        #     if not len(value):
        #         continue
        #     assert len(value) == 1
        #     value = value[0]

        # if isinstance(value, str) and value.endswith('。') and slot == '特点':
        #     value = value[:-1]
        #
        # value = process_string(value)
        #
        # translated_item[value_mapping.zh2en_SLOT_MAP[slot]] = value

        # if isinstance(value, str):
        #     translated_item[slot] = ''.join(value.split())
        # else:
        #     translated_item[slot] = value

        translated_db_info.append(translated_item)

    json.dump(
        translated_db_info, open(os.path.join(args.output_db_dir, domain + '_zh.json'), 'w'), ensure_ascii=False, indent=2
    )

print(dict(sorted(dicti.items(), key=lambda item: item[0])))



# Copyright 2022 The Board of Trustees of the Leland Stanford Junior University
#
# Author: Mehrad Moradshahi <mehrad@cs.stanford.edu>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dialogues',
    version='0.0.3',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': ['dialogues=dialogues.__main__:main'],
    },
    package_data={
        'dialogues': [
            'bitod/src/knowledgebase/apis/*.json',
            'bitod/src/knowledgebase/dbs/*.json',
            'bitod/src/knowledgebase/mappings/*.json',
            'bitod/src/knowledgebase/mappings/*.dot',
            'bitod/src/templates/files/*/*',
            'risawoz/src/db/*.json',
            'risawoz/src/knowledgebase/mappings/*.json',
        ]
    },
    license='BSD-3-Clause',
    author="Mehrad Moradshahi",
    author_email="mehrad@cs.stanford.edu",
    description="This package provides a unified interface to several dialogue benchmarks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mehrad0711/dialogues",
    install_requires=[
        'pydot==1.4.2',
        'pymongo==3.11.2',
        'dnspython==2.1.0',
        'word2number==1.1',
        'dictdiffer~=0.9',
        'networkx==2.5',
        'sacrebleu>=1,<3',
        'datasets~=2.1',
    ],
)



import argparse
import json
from collections import defaultdict

from dialogues.risawoz.src.knowledgebase.api import process_string

parser = argparse.ArgumentParser()

parser.add_argument('--input_file', default='dialogues/risawoz/src/knowledgebase/mappings/zh2en_alignment.json')

args = parser.parse_args()


zh2en_value = defaultdict(lambda: defaultdict(dict))

with open(args.input_file) as fin:
    zh2en_alignment = json.load(fin)
    for domain, items in zh2en_alignment.items():
        for slot, values in items.items():
            for zh_val, en_vals in values.items():
                value = process_string(zh_val)
                value = value.replace('，', ',')
                value = value.replace('：', ':')
                zh2en_value[domain][slot][value] = en_vals

with open(args.input_file.replace('zh2en_alignment.json', 'zh2en_alignment_new.json'), 'w') as fout:
    json.dump(zh2en_value, fout, ensure_ascii=False, indent=4)
